import configparser
import glob
import os
import platform
import re
import sys
from dataclasses import dataclass

from .os_specific import OsSpecific
from .svc import svc


# --------------------
## perform the do_check operation.
# checks all installation and environment
# Do not use any non-built in modules
# Do not use out/ directory
class DoCheck:
    ## list of tools needed; max 3 nodes of the version string
    # if cmd is None, then a specific function must be used
    valid_versions = {
        'python': {
            'techs': ['python', 'cpp', 'arduino'],
            'cmd': None,
            'valid': {
                'ubuntu': ['3.8', '3.9', '3.10'],
                'macos': ['3.9', '3.10'],
                'win': ['3.10'],
                'rpi': [],
            },
            'verbose': True,
        },
        'tkinter': {
            'techs': ['python'],
            'cmd': None,
            'valid': {
                'ubuntu': ['8.6'],
                'macos': ['8.6'],
                'win': ['8.6'],
                'rpi': [],
            },
            'verbose': True,
        },
        'doxygen': {
            'techs': ['python', 'cpp', 'arduino'],
            'cmd': 'doxygen -v',
            'valid': {
                'ubuntu': ['1.8.17', '1.9.1', '1.9.5', '1.10.0'],
                'macos': ['1.9.8', '1.10.0'],
                'win': ['1.10.0', '1.12.0'],
                'rpi': [],
            },
            'verbose': False,
        },
        'libreoffice': {
            'techs': ['python'],
            'cmd': 'libreoffice --version',
            'macos:cmd': 'soffice --version',
            'win:cmd': '"C:\\Program Files\\LibreOffice\\program\\soffice.com" --version',
            'valid': {
                'ubuntu': ['24.2.5', '24.2.6', '24.8.2'],
                'macos': ['7.6.4', '24.2.0', '24.8.2'],
                'win': ['24.2.0'],
                'rpi': [],
            },
            'verbose': False,
        },
        'graphviz': {
            'techs': ['python'],
            'cmd': 'dot -V',
            'valid': {
                'ubuntu': ['2.43.0'],
                'macos': ['9.0.0', '12.1.2'],
                'win': ['9.0.0', '10.0.1', '12.1.2'],
                'rpi': [],
            },
            'verbose': False,
        },
        # deprecated: cmake not used
        # 'cmake': {
        #     'techs': ['cpp', 'arduino'],
        #     'cmd': 'cmake --version',
        #     'valid': {
        #         'ubuntu': ['3.16.3', '3.22.1', '3.25.1'],
        #         'macos': ['3.30.5'],
        #         'win': [],
        #         'rpi': [],
        #     },
        #     'verbose': False,
        # },
        # deprecated: ninja not used
        # 'ninja': {
        #     'techs': ['cpp', 'arduino'],
        #     'cmd': 'ninja --version',
        #     'valid': {
        #         'ubuntu': ['1.10.0', '1.10.1'],
        #         'macos': ['1.12.1'],
        #         'win': [],
        #         'rpi': [],
        #     },
        #     'verbose': False,
        # },
        'gcc': {
            'techs': ['cpp'],
            'cmd': 'gcc --version',
            'valid': {
                'ubuntu': ['9.4.0', '11.4.0', '13.2.0'],
                'macos': ['15.0.0'],
                'win': ['14.2.0'],
                'rpi': [],
            },
            'verbose': False,
        },
        'clang-tidy': {
            'techs': ['cpp', 'arduino'],
            'cmd': 'clang-tidy --version',
            'valid': {
                'ubuntu': ['10.0.0', '14.0.0', '18.1.3'],
                'macos': ['19.1.1'],
                'win': ['18.1.8'],
                'rpi': [],
            },
            'verbose': False,
        },
        'gcovr': {
            'techs': ['cpp', 'arduino'],
            'cmd': 'gcovr --version',
            'valid': {
                'ubuntu': ['4.2', '5.0', '7.2', '7.0', '8.2'],
                'macos': ['8.2'],
                'win': ['8.2'],
                'rpi': [],
            },
            'verbose': False,
        },
        'avr-gcc': {
            'techs': ['arduino'],
            'cmd': 'avr-gcc --version',
            'valid': {
                'ubuntu': ['5.4.0', '7.3.0'],
                'macos': [],
                'win': ['14.2.0'],
                'rpi': [],
            },
            'verbose': False,
        },
        'avrdude': {
            'techs': ['arduino'],
            'cmd': 'avrdude -?',
            'valid': {
                'ubuntu': ['6.3', '7.1'],
                'macos': [],
                'win': ['6.3'],
                'rpi': [],
            },
            'verbose': False,
        },
        'avr-gdb': {
            'techs': ['arduino'],
            'cmd': 'avr-gdb --version',
            'win:cmd': 'echo "0.0"',  # doesn't install in windows
            'valid': {
                'ubuntu': ['8.1.0', '10.1.90', '15.0.50'],
                'macos': [],
                'win': ['0.0'],
                'rpi': [],
            },
            'verbose': False,
        },
        'ruby': {
            'techs': ['cpp'],
            'cmd': 'ruby --version',
            'valid': {
                'ubuntu': ['3.0.2', '3.2.3'],
                'macos': ['3.3.5'],
                'win': ['3.3.6'],
                'rpi': [],
            },
            'verbose': False,
        },
        'swig': {
            'techs': ['cpp'],
            'cmd': 'swig --version',
            'win:cmd': 'swig -version',
            'valid': {
                'ubuntu': ['4.3.0'],
                'macos': ['4.2.1'],
                'win': ['4.2.1'],
                'rpi': [],
            },
            'verbose': False,
        }
    }

    # --------------------
    ## constructor
    def __init__(self):
        ## holds additional scripts to check for permissions
        self._scripts = []

    # --------------------
    ## do_check mainline.
    #
    # @return None
    def run(self):
        svc.log.highlight(f'{svc.gbl.tag}: starting...')

        self._check_ostype()
        # python is needed by arduino and cpp projects as well
        self._check_python_version()

        # make sure mod names are correct first
        self._check_mod_names()

        # TODO delete if not needed
        # ensure version.json and build_info are created
        # if svc.gbl.rc == 0 and dogen == 'dogen':
        #     # only if modnames are ok
        #     svc.gen_files.all(dogen=dogen)

        self._gen_license()

        # check everything else
        self._check_versions()
        self._check_bash_scripts()
        self._check_common()

        if svc.cfg.mod_tech == 'python':
            self._check_publish_files()
            self._check_pypi()
            self._check_tkinter()

        if svc.cfg.mod_tech == 'arduino':
            self._check_arduino()

        svc.gbl.rc = svc.gbl.rc
        svc.gbl.overallrc += svc.gbl.rc

    # --------------------
    ## check the OS name and platform are recognized
    #
    # @return None
    def _check_ostype(self):
        svc.log.line(f'os_name : {os.name}:{OsSpecific.os_name}')

        svc.log.line(f'system  : {platform.system()}')
        # Linux: Linux
        # Mac: Darwin
        # Windows: Windows
        # RPI: ??

        svc.log.line(f'platform: {sys.platform}')
        # Linux: linux (lower case)
        # Mac: ??
        # Win MSYS2: msys (need to confirm)
        # Win MING : mingw64
        # WIN WSL  : linux2
        # RPI: ??

    # --------------------
    ## check the python version
    #
    # @return True if a valid version, False otherwise
    def _check_python_version(self):
        svc.log.highlight('python version:')
        version = sys.version.replace('\n', ' ')
        svc.log.line(f'version : {version}')
        svc.log.line(f'info    : {sys.version_info}')

        version = f'{sys.version_info.major}.{sys.version_info.minor}'
        self._check_tool_version2('python', version, True)

    # --------------------
    ## check all tool versions
    #
    # @return None
    def _check_versions(self):
        msgs = []
        ok = True
        for tool, item in self.valid_versions.items():
            if svc.cfg.mod_tech not in item['techs']:
                # uncomment to debug
                # svc.log.dbg(f'check_versions: tech:{svc.cfg.mod_tech} skipping {tool}')
                continue

            # check if there is an os_specific cmd to run
            os_cmd = f'{OsSpecific.os_name}:cmd'
            if os_cmd in item and item[os_cmd] is not None:
                ok = ok and self._check_tool_version(msgs, tool, item, item[os_cmd])
                continue

            # no os_specific one, use the generic one
            if item['cmd'] is not None:
                ok = ok and self._check_tool_version(msgs, tool, item, item['cmd'])
                continue

        if ok:
            svc.log.ok('all tool versions')
        else:
            for msg in msgs:
                svc.gbl.rc += 1
                svc.log.err(msg)

    # --------------------
    ## check the version for the given tool
    #
    # @param msgs    list of error messages to display
    # @param key     the name of the tool
    # @param item    tool information from valid_versions
    # @param cmd     the command to use
    # @return None
    def _check_tool_version(self, msgs, key, item, cmd):
        rc, out = svc.utils_ps.run_cmd(cmd)
        ok = True
        if rc != 0:
            ok = False
            msgs.append(f'{key} version failed, rc:{rc}, out:"{out}"')
        else:
            # note: 3rd spot is optional
            m = re.search(r'(\d+\.\d+(\.\d+)?)', out)
            if m:
                version = m.group(1)
            else:
                version = out
            self._check_tool_version2(key, version, item['verbose'])

        return ok

    # --------------------
    ## checks version on the given tool
    #
    # @param tool_name    the tool to check
    # @param act_version  the actual version of the tool
    # @param verbose      flag to indicate whether to log passing chacks or not
    # @return None
    def _check_tool_version2(self, tool_name, act_version, verbose):
        valid = self.valid_versions[tool_name]['valid'][OsSpecific.os_name]
        # version is sometimes a float
        if str(act_version) in valid:
            if verbose:
                svc.log.ok(f'{tool_name} {act_version}')
        else:
            svc.gbl.rc += 1
            svc.log.err(f'{tool_name} valid versions {valid}, actual: {act_version}')

    # --------------------
    ## check for the existence of $HOME/.pypirc and its content
    #
    # @return None
    def _check_pypi(self):
        home = os.path.expanduser('~')
        path = os.path.join(home, '.pypirc')
        exists = os.path.isfile(path)
        msg = f'{".pypirc exists": <15}: {exists}'
        svc.log.check(exists, msg)

        if exists:
            self._check_pypi_content(path)
        else:
            svc.gbl.rc += 1

    # --------------------
    ## check for the existence of $HOME/.pypirc and its content
    #
    # @return None
    def _check_tkinter(self):
        # >>> import tkinter
        # >>> tkinter.TkVersion
        # 8.6
        import importlib.util
        tkmod = importlib.util.find_spec('tkinter')
        if tkmod is None:
            svc.log.warn('tkinter not installed')
        else:
            import tkinter
            version = tkinter.TkVersion
            self._check_tool_version2('tkinter', version, True)

    # --------------------
    ## check the content of the .pypirc file
    #
    # @param path the path to the file
    # @return None
    def _check_pypi_content(self, path):
        config = configparser.ConfigParser()
        config.read(path)
        msg = ''
        ok = 'pypi' in config.sections()
        if not ok:
            svc.gbl.rc += 1
            msg += 'missing "pypi" section'
            svc.log.err(msg)
            return

        msgs = []
        # check username
        if 'username' not in config['pypi']:
            msgs.append('missing pypi.username')
            ok = False
        elif config['pypi']['username'] != '__token__':
            msgs.append('pypi.username should be "__token__"')
            ok = False

        # check password
        if 'password' not in config['pypi']:
            msgs.append('missing pypi.password')
            ok = False

        svc.log.check_all(ok, f'{".pypirc content": <15}', msgs)

    # --------------------
    ## check permissions on the executable scripts
    #  * all scripts that start with do_*
    #  * any given in self._scripts
    #
    # @return None
    def _check_bash_scripts(self):
        scripts = glob.glob('do*')
        self._scripts.extend(scripts)
        ok = True
        for path in self._scripts:
            if not os.path.isfile(path):
                continue
            ok = ok and self._check_script(path)
        if ok:
            svc.log.ok('execute permissions')

    # --------------------
    ## checks the permissions for the given script
    #
    # @param path   the script to check
    # @return None
    def _check_script(self, path):
        ok = os.access(path, os.X_OK)
        if not ok:
            svc.gbl.rc += 1
            msg = f'execute permission missing: {path}'
            svc.log.err(msg)
        return ok

    # --------------------
    ## check values against xplat.cfg content
    #
    # @return None
    def _check_common(self):
        self._check_doxyfile()
        self._check_gitignore()
        self._check_gitconfig()
        self._check_license()
        self._check_requirements_txt()
        self._check_home()

    # --------------------
    ## check module names
    #
    # @return None
    def _check_mod_names(self):
        # check for default mod_names
        if svc.cfg.is_module:
            svc.log.highlight(f'{svc.cfg.mod_name} is a module')
        else:
            svc.log.highlight(f'{svc.cfg.mod_name} is not a module')

        if svc.cfg.mod_name == 'module-name':
            svc.gbl.rc += 1
            svc.log.err(f'{svc.cfg.mod_name} must be changed to correct name')
        if svc.cfg.mod_dir_name == 'module_name':
            svc.gbl.rc += 1
            svc.log.err(f'{svc.cfg.mod_dir_name} must be changed to correct dir name')

        # check mod names are correct format
        if not svc.cfg.mod_name:
            svc.gbl.rc += 1
            svc.log.err('mod_name in xplat.cfg not set')
        elif '_' in svc.cfg.mod_name:
            svc.gbl.rc += 1
            svc.log.err(f'mod_name should contain "-", not "_": {svc.cfg.mod_name}')
        else:
            svc.log.ok(f'{"mod_name": <18}: {svc.cfg.mod_name}')

        if svc.cfg.mod_dir_name == '':
            svc.gbl.rc += 1
            svc.log.err('mod_dir_name in xplat.cfg not set')
        elif '-' in svc.cfg.mod_dir_name:
            svc.gbl.rc += 1
            svc.log.err(f'mod_dir_name should contain "_", not "-": {svc.cfg.mod_dir_name}')
        else:
            svc.log.ok(f'{"mod_dir_name": <18}: {svc.cfg.mod_dir_name}')

    # --------------------
    ## check Doxyfile
    #
    # @return None
    def _check_doxyfile(self):
        tag = 'Doxyfile'
        fname = os.path.join(svc.utils_fs.root_dir, 'Doxyfile')
        if not os.path.isfile(fname):
            svc.log.warn(f'{tag}: Doxyfile not found, skipping check')
            return

        @dataclass
        class state:  # pylint: disable=invalid-name
            found_version = False
            found_name = False
            found_exclude = False
            ok = True

        with open(fname, 'r', encoding='UTF-8') as fp:
            for line in fp:
                line = line.strip()

                self._doxyfile_project_name(tag, line, state)
                self._doxyfile_project_number(tag, line, state)
                self._doxyfile_excludes(tag, line, state)

        if not state.found_version:
            svc.gbl.rc += 1
            svc.log.err(f'{tag}: "PROJECT_NUMBER" line not found')
            state.ok = False

        if not state.found_name:
            svc.gbl.rc += 1
            svc.log.err(f'{tag}: "PROJECT_NAME" line not found')
            state.ok = False

        if not state.found_exclude:
            # even if the module doesn't use build_info, Doxyfile can exclude it
            svc.gbl.rc += 1
            svc.log.err(f'{tag}: "EXCLUDE" build_info line not found')
            state.ok = False

        if state.ok:
            svc.log.ok(f'{tag}')

    # --------------------
    ## check Doxyfile project name
    # e.g. PROJECT_NAME = "gui-api-tkinter Module" or
    # for apps: PROJECT_NAME = "gui-api-tkinter App"
    #
    # @param tag    the logging tag
    # @param line   the doxyfile line to check
    # @param state  (dict) holds the current state of the checks
    # @return None
    def _doxyfile_project_name(self, tag, line, state):
        m = re.search(r'PROJECT_NAME\s*=\s*(".+")$', line)
        if not m:
            return

        # found it
        if svc.cfg.is_module:
            exp_name = 'Module'
        else:
            exp_name = 'App'

        state.found_name = True
        if m.group(1) != f'"{svc.cfg.mod_name} {exp_name}"':
            svc.gbl.rc += 1
            svc.log.err(f'{tag}: "PROJECT_NAME" line does not match "{svc.cfg.mod_name} {exp_name}", '
                        f'actual: {line}')
            state.ok = False

    # --------------------
    ## check Doxyfile project version
    # e.g. PROJECT_NUMBER = 0.0.1
    #
    # @param tag    the logging tag
    # @param line   the doxyfile line to check
    # @param state  (dict) holds the current state of the checks
    # @return None
    def _doxyfile_project_number(self, tag, line, state):
        m = re.search(r'PROJECT_NUMBER\s*=\s*(.+)\s*$', line)
        if not m:
            return

        # found it
        state.found_version = True
        if m.group(1) != f'v{svc.cfg.version}':
            svc.gbl.rc += 1
            svc.log.err(f'{tag}: "PROJECT_NUMBER" version does not match version.json: {svc.cfg.version}, '
                        f'actual: {line}')
            state.ok = False

    # --------------------
    ## check Doxyfile excludes
    # e.g. EXCLUDE += ./.../lib/build_info.py or
    # for apps EXCLUDE += ./lib/build_info.py
    #
    # @param tag    the logging tag
    # @param line   the doxyfile line to check
    # @param state  (dict) holds the current state of the checks
    # @return None
    def _doxyfile_excludes(self, tag, line, state):
        if OsSpecific.os_name == 'win':
            slash = '\\\\'
        else:
            slash = '/'
        if svc.cfg.mod_tech == 'python':
            m = re.search(fr'EXCLUDE\s*\+=\s*(.+){slash}build_info\.py', line)
        else:
            m = re.search(fr'EXCLUDE\s*\+=\s*(.+){slash}build_info\.txt', line)
        if not m:
            return

        if svc.cfg.is_module:
            if svc.cfg.mod_tech == 'python':
                exp_dir = f'.{os.sep}{svc.cfg.mod_dir_name}{os.sep}lib'
            else:
                exp_dir = f'.{os.sep}lib'
        else:
            if svc.cfg.mod_tech == 'python':
                exp_dir = f'.{os.sep}lib'
            else:
                exp_dir = f'.{os.sep}src'

        # found it
        state.found_exclude = True
        if m.group(1) != exp_dir:
            svc.gbl.rc += 1
            if svc.cfg.mod_tech == 'python':
                svc.log.err(f'{tag}: "EXCLUDE" build_info line does not match {exp_dir}/build_info.py, '
                            f'actual: {line}')
            else:
                svc.log.err(f'{tag}: "EXCLUDE" build_info line does not match {exp_dir}/build_info.txt, '
                            f'actual: {line}')
            state.ok = False

    # --------------------
    ## check .gitignore file
    #
    # @return None
    def _check_gitignore(self):
        tag = '.gitignore'

        fname = self._get_gitignore_path()
        if fname is None:
            return

        if not os.path.isfile(fname):
            svc.log.err(f'{tag}: file not found: {fname}')
            return

        @dataclass
        class state:  # pylint: disable=invalid-name
            found_buildinfo = False
            ok = True

        with open(fname, 'r', encoding='UTF-8') as fp:
            for line in fp:
                line = line.strip()

                if svc.cfg.mod_tech in ['cpp', 'arduino']:
                    if re.search(r'^\s*/build_info.txt\s*$', line):
                        state.found_buildinfo = True
                        continue
                else:  # python
                    if re.search(r'^\s*/build_info.py\s*$', line):
                        state.found_buildinfo = True
                        continue

        if not state.found_buildinfo:
            svc.gbl.rc += 1
            svc.log.err(f'{tag}: build_info line not found')
            state.ok = False

        if state.ok:
            svc.log.ok(f'{tag}')

    # --------------------
    ## get the path to the local gitignore file
    #
    # @return path to gitignore
    def _get_gitignore_path(self):
        if svc.cfg.mod_tech == 'cpp':
            if svc.cfg.is_module:
                path = os.path.join(svc.utils_fs.root_dir, 'lib', '.gitignore')
            else:
                path = os.path.join(svc.utils_fs.root_dir, 'src', '.gitignore')
        elif svc.cfg.mod_tech == 'arduino':
            path = os.path.join(svc.utils_fs.root_dir, 'src', '.gitignore')
        else:  # python
            if svc.cfg.is_module:
                mod_dir = os.path.join(svc.utils_fs.root_dir, svc.cfg.mod_dir_name)
                if not os.path.isdir(mod_dir):
                    svc.log.warn(f'check_gitignore: dir does not exist: {mod_dir}')
                    svc.gbl.rc = 1
                    return None

                path = os.path.join(str(mod_dir), 'lib', '.gitignore')
            else:
                path = os.path.join(svc.utils_fs.root_dir, 'lib', '.gitignore')

        return path

    # --------------------
    ## check .gitignore file
    #
    # @return None
    def _check_gitconfig(self):
        tag = '.gitconfig'
        path = os.path.join(os.path.expanduser('~'), '.gitconfig')

        ok = True
        if not os.path.isfile(path):
            svc.gbl.rc += 1
            svc.log.err(f'{tag} cannot find {path}')
            return

        config = configparser.ConfigParser()
        config.read(path)
        ok = ok and self._check_gitconfig_section(tag, config, 'pull', 'rebase', 'false')
        ok = ok and self._check_gitconfig_section(tag, config, 'push', 'autoSetupRemote', 'true')

        if ok:
            svc.log.ok(f'{tag}')

    # --------------------
    ## check a section in the .gitignore file
    #
    # @param tag      the logging tag
    # @param config   reference to the configparser
    # @param section  the section name
    # @param name     the name of the line
    # @param value    the value of the line
    # @return return True if everything is okay in the section, false otherwise
    def _check_gitconfig_section(self, tag, config, section, name, value):
        ok = True
        if section not in config.sections():
            svc.gbl.rc += 1
            ok = False
            svc.log.err(f'{tag} cannot find "{section}" section')
        elif name not in config[section]:
            svc.gbl.rc += 1
            ok = False
            svc.log.err(f'{tag} cannot find "{name}" in "{section}" section')
        elif config[section][name] != value:
            svc.gbl.rc += 1
            ok = False
            svc.log.err(f'{tag} {section}.{name} should be "{value}", actual: {config[section][name]}')
        return ok

    # --------------------
    ## check license file
    #
    # @return None
    def _check_license(self):
        # * check license, for "Copyright line"
        tag = 'LICENSE.txt'
        fname = 'LICENSE.txt'
        path = os.path.join(svc.utils_fs.root_dir, fname)

        if svc.cfg.is_module and not os.path.isfile(path):
            svc.gbl.rc += 1
            svc.log.err(f'{tag} modules must have a LICENSE.txt file')
            return

        if not svc.cfg.is_module and not os.path.isfile(path):
            # apps may or may not have a LICENSE.txt file
            return

        with open(path, 'r', encoding='UTF-8') as fp:
            line1 = fp.readline().strip()
            line2 = fp.readline().strip()
            line3 = fp.readline().strip()
            line4 = fp.readline().strip()
            line5 = fp.readline().strip()
            line6 = fp.readline().strip()

            ok = True
            # MIT License
            if line1 != f'{svc.cfg.license} License':
                svc.gbl.rc += 1
                ok = False
                svc.log.err(f'{fname}: line1 invalid license line: "{line1}"')

            if line2:
                svc.gbl.rc += 1
                ok = False
                svc.log.err(f'{fname}: line2 should be blank: "{line2}"')

            # Copyright (c) since (etc.)
            m = re.search(r'^Copyright \(c\) since \d{4}: ', line3)
            if not m:
                svc.gbl.rc += 1
                ok = False
                svc.log.err(f'{fname}: line3 invalid copyright line: "{line3}"')

            # a one-line description
            m = re.search(r'^description: ', line4)
            if not m:
                svc.gbl.rc += 1
                ok = False
                svc.log.err(f'{fname}: line4 should contain a description: "{line4}"')

            m = re.search(r'^author: .* email: .*$', line5)
            if not m:
                svc.gbl.rc += 1
                ok = False
                svc.log.err(f'{fname}: line5 should have author and email: "{line5}"')

            if line6:
                svc.gbl.rc += 1
                ok = False
                svc.log.err(f'{fname}: line6 should be blank: "{line6}"')

            if ok:
                svc.log.ok(f'{fname}')

    # --------------------
    ## check files used in module publish.
    # * if module, must exist
    # * if app, must not exist
    #
    # @return None
    def _check_publish_files(self):
        tag = 'publish'
        for fname in ['do_publish']:
            path = os.path.join(svc.utils_fs.root_dir, fname)
            if svc.cfg.is_module and not os.path.isfile(path):
                svc.gbl.rc += 1
                svc.log.err(f'{tag}" modules must have a "{fname}" file')
            elif not svc.cfg.is_module and os.path.isfile(path):
                svc.gbl.rc += 1
                svc.log.err(f'{tag}: apps should not have a "{fname}" file')

    # --------------------
    ## check requirements file
    #
    # @return None
    def _check_requirements_txt(self):
        ok = True
        tag = 'requirements.txt'
        path = os.path.join(svc.utils_fs.root_dir, 'tools', 'requirements.txt')

        if not os.path.isfile(path):
            svc.gbl.rc += 1
            svc.log.err(f'{tag} does not exist: {path}')
            return

        if ok:
            svc.log.ok(f'{tag}')

    # --------------------
    ## check definition of HOME
    #
    # @return None
    def _check_home(self):
        py_home = os.path.expanduser('~')
        if OsSpecific.os_name == 'win':
            py_home2 = py_home.replace('\\', '/')
            py_home2 = py_home2.replace('C:', '/c')
            _, bash_home = svc.utils_ps.run_cmd('echo ~', msys2=True)
        else:
            py_home2 = py_home
            _, bash_home = svc.utils_ps.run_cmd('echo ~')

        if py_home2 == bash_home:
            svc.log.ok(f'home directory: {bash_home}')
        else:
            svc.log.warn('home directory (~) and pyhon home are defined differently:')
            svc.log.warn(f'   python: {py_home}')
            svc.log.warn(f'   bash  : {bash_home}')

    # --------------------
    ## check arduino libraries etc.
    #
    # @return None
    def _check_arduino(self):
        tag = 'arduino'

        if OsSpecific.os_name == 'macos':
            path = os.path.expanduser('~/Library/Arduino15')
        elif OsSpecific.os_name == 'win':
            path = os.path.expanduser('~/AppData/Local/Arduino15')
        else:
            path = os.path.join(os.sep, 'usr', 'share', 'arduino')
        if not os.path.isdir(path):
            svc.gbl.rc += 1
            svc.log.err(f'{tag}: missing {path}, run "sudo apt install arduino"')
        elif OsSpecific.os_name == 'ubuntu':
            # /usr/share/arduino/lib/version.txt
            version = os.path.join(path, 'lib', 'version.txt')
            with open(version, 'r', encoding='utf-8') as fp:
                line = fp.readline()
                svc.log.ok(f'{tag}: found lib version: {line.strip()}')

        if OsSpecific.os_name == 'ubuntu':
            path = os.path.join(os.sep, 'usr', 'lib', 'gcc', 'avr')
            if not os.path.isdir(path):
                svc.gbl.rc += 1
                svc.log.err(f'{tag}: missing {path}, run "sudo apt install avr-libc"')
            else:
                dirs = os.listdir(path)
                for subdir in dirs:
                    svc.log.ok(f'{tag}: found avr lib version: {subdir}')

    # --------------------
    ## gen licence file using template
    #
    # @return None
    def _gen_license(self):  # pylint: disable=too-many-branches
        svc.utils_fs.safe_delete_file('LICENSE.txt')

        # generate the LICENSE.txt file
        src = os.path.join(svc.utils_fs.root_dir, 'tools', 'xplat_utils', 'license_template.txt')
        dst = os.path.join(svc.utils_fs.root_dir, 'LICENSE.txt')
        with open(dst, 'w', encoding='utf-8', newline='\n') as fp_dst:
            with open(src, 'r', encoding='utf-8') as fp_src:
                for line in fp_src:
                    while True:
                        m = re.search(r'\$\$([^$]+)\$\$', line)
                        if not m:
                            # uncomment to debug
                            # log.dbg(f'line: {line.strip()}')
                            fp_dst.write(line)
                            break

                        vrbl = m.group(1)
                        if vrbl == 'module-name':
                            val = getattr(svc.cfg, 'mod_name')
                        elif vrbl == 'lic-tech':
                            if svc.cfg.mod_tech == 'python':
                                val = 'Python'
                            elif svc.cfg.mod_tech == 'cpp':
                                val = 'C/C++'
                            elif svc.cfg.mod_tech == 'arduino':
                                val = 'Arduino'
                            else:
                                svc.log.warn(f'gen_license: unknown mod_tech found: {svc.cfg.mod_tech}')
                                val = svc.cfg.mod_tech
                        elif vrbl == 'lic-proj-type':
                            if svc.cfg.is_module:
                                if svc.cfg.mod_tech == 'python':
                                    val = 'Module'
                                else:
                                    val = 'Library'
                            else:
                                val = 'App'
                        elif vrbl == 'lic-year':
                            val = getattr(svc.cfg, 'lic_year', None)
                            if val is None:
                                val = '2024'
                                svc.log.warn(f'gen_license: unknown lic_year, using {val}')
                        elif vrbl == 'lic-desc':
                            val = getattr(svc.cfg, 'lic_desc', None)
                            if val is None:
                                val = svc.cfg.do_doc.doxy_desc
                        else:
                            val = getattr(svc.cfg, vrbl)
                        # uncomment to debug
                        # log.dbg(f'grp:{m.group(1)} line:{line.strip()}')
                        line = line.replace(f'$${vrbl}$$', val)

        svc.gbl.overallrc += svc.gbl.rc



import json
import os

from .os_specific import OsSpecific
from .svc import svc


# --------------------
## perform the do_ver_info operation
class DoVerInfo:
    # --------------------
    ## constructor
    def __init__(self):
        ## the list of versions tested so far
        self._test_versions = {}

    # --------------------
    ## do_ver_info mainline.
    #
    # @return None
    def run(self):
        svc.log.highlight(f'{svc.gbl.tag}: starting save...')

        self._gen_version_info()
        self._gen_markdown()

    # --------------------
    ## generate or udpate version info json file
    #
    # @return None
    def _gen_version_info(self):
        svc.log.line(f'{svc.gbl.tag}: generating version_info')

        # this location is hardcoded
        path = os.path.join('tools', 'version_info.json')

        if os.path.isfile(path):
            # load existing data
            with open(path, 'r', encoding='utf-8') as fp:
                self._test_versions = json.load(fp)
        else:
            # first time generating the file
            # Note: these must match the tags in OsSpecific
            self._test_versions = {
                'ubuntu': {},
                'win': {},
                'macos': {},
                'rpi': {},
            }

        # get version info
        info = f'{OsSpecific.os_version}, {OsSpecific.python_version}'
        self._test_versions[OsSpecific.os_name][info] = 1

        # save it
        with open(path, 'w', encoding='utf-8', newline='\n') as fp:
            json.dump(self._test_versions, fp, indent=4)

    # --------------------
    ##  generates the markdown file
    def _gen_markdown(self):
        svc.log.line(f'{svc.gbl.tag}: generating markdown')

        # TODO get location from .cfg
        # generate list of versions for each platform
        path = os.path.join('doc', 'version_info.md')
        with open(path, 'w', encoding='utf-8', newline='\n') as fp:
            for platform in sorted(self._test_versions.keys()):
                for version in sorted(self._test_versions[platform].keys()):
                    fp.write(f'{version}\n')



import sys


# -------------------
## Holds all info for logging debug lines
class UtilsLogger:
    ## flag to log to stdout or not
    verbose = True
    ## for UT only
    ut_mode = False
    ## for UT only
    ut_lines = []

    # --------------------
    ## log a message. Use ok() or err() as appropriate.
    #
    # @param ok      the check state
    # @param msg     the message to print
    # @return None
    @staticmethod
    def check(ok, msg):
        if ok:
            UtilsLogger.ok(msg)
        else:
            UtilsLogger.err(msg)

    # --------------------
    ## log a series of messages. Use ok() or err() as appropriate.
    #
    # @param ok      the check state
    # @param title   the line indicating what the check is about
    # @param msgs    individual list of lines to print
    # @return None
    @staticmethod
    def check_all(ok, title, msgs):
        UtilsLogger.check(ok, f'{title}: {ok}')
        for msg in msgs:
            UtilsLogger.check(ok, f'   - {msg}')

    # -------------------
    ## write a "====" line with the given message
    #
    # @param msg     the message to write
    # @return None
    @staticmethod
    def start(msg):
        UtilsLogger._write_line('====', msg)

    # -------------------
    ## write a "line" line with the given message
    #
    # @param msg     the message to write
    # @return None
    @staticmethod
    def line(msg):
        UtilsLogger._write_line(' ', msg)

    # -------------------
    ## write a "highlight" line with the given message
    #
    # @param msg     the message to write
    # @return None
    @staticmethod
    def highlight(msg):
        UtilsLogger._write_line('--->', msg)

    # -------------------
    ## write a "ok" line with the given message
    #
    # @param msg     the message to write
    # @return None
    @staticmethod
    def ok(msg):
        UtilsLogger._write_line('OK', msg)

    # -------------------
    ## write a "err" line with the given message
    #
    # @param msg     the message to write
    # @return None
    @staticmethod
    def err(msg):
        UtilsLogger._write_line('ERR', msg, always_print=True)

    # -------------------
    ## write a "bug" line with the given message
    #
    # @param msg     the message to write
    # @return None
    @staticmethod
    def bug(msg):
        UtilsLogger._write_line('BUG', msg, always_print=True)

    # -------------------
    ## write an output line with the given message
    #
    # @param msg     the message to write
    # @param lineno  (optional) the current line number for each line printed
    # @return None
    @staticmethod
    def output(msg, lineno=None):
        if lineno is None:
            tag = ' --    '
        else:
            tag = f' --{lineno: >3}]'
        UtilsLogger._write_line(tag, msg)

    # -------------------
    ## write a list of lines using output()
    #
    # @param lines   the lines to write
    # @return None
    @staticmethod
    def num_output(lines):
        lineno = 0
        for line in lines:
            lineno += 1
            UtilsLogger.output(line, lineno=lineno)

    # -------------------
    ## write a "warn" line with the given message
    #
    # @param msg     the message to write
    # @return None
    @staticmethod
    def warn(msg):
        UtilsLogger._write_line('WARN', msg)

    # -------------------
    ## write a "err" line with the given message
    #
    # @param msg     the message to write
    # @return None
    @staticmethod
    def dbg(msg):
        UtilsLogger._write_line('DBG', msg)

    # -------------------
    ## write a raw line (no tag) with the given message
    #
    # @param msg     the message to write
    # @return None
    @staticmethod
    def raw(msg):
        UtilsLogger._write_line(None, msg)

    # -------------------
    ## write the given line to stdout
    #
    # @param tag           the prefix tag
    # @param msg           the message to write
    # @param always_print  print the message even if verbose is False
    # @return None
    @staticmethod
    def _write_line(tag, msg, always_print=False):
        if not UtilsLogger.verbose and not always_print:
            return

        # TODO add ability to optionally save to file

        if tag is None:
            line = msg
        else:
            line = f'{tag: <4} {msg}'

        if UtilsLogger.ut_mode:
            UtilsLogger.ut_lines.append(line)
        else:
            print(line)  # print okay
            sys.stdout.flush()



import getpass
import os
import platform
import subprocess
import sys

from .svc import svc


# -------------------
## runs OS specific commands
# there are four recognized OS:
#  * Ubuntu
#  * Mac
#  * Windows
#  * RPi
class OsSpecific:
    ## holds the OS name
    os_name = 'unknown'
    ## list of valid OS
    os_valid = ['ubuntu', 'macos', 'win', 'rpi']
    ## holds the OS version info
    os_version = 'unknown'
    ## holds the python version
    python_version = 'unknown'

    ## holds a reference to the implemenation class for the current OS
    impl = None

    # -------------------
    ## implements the Windows specific commands
    class Win:
        # -------------------
        ## constructor
        def __init__(self):
            pass

        # -------------------
        ## simple OS name
        #
        # @return string indicating OS
        def os_name(self):
            return 'win'

        # -------------------
        ## holds OS information
        #
        # @return string indicating OS info
        def os_version(self):
            return f'win32 {platform.system()} {platform.release()}'

        # # -------------------
        # ## returns the given command line
        # # Windows implementation needs the un-parsed command line
        # #
        # # @param cmdline the command line to parse
        # # @return the given cmdline
        # def parse_command_line(self, cmdline):
        #     return 'cmd /C ' + cmdline

        # # -------------------
        # ## returns a preexec_fn for Popen calls
        # # not used in Windows implementation
        # #
        # # @return None
        # def preexec_fn(self):
        #     return None

        # # -------------------
        # ## returns the package manager
        # # Windows does not have a package manager
        # #
        # # @return None
        # def package_manager(self):
        #     return None

        # # -------------------
        # ## kill the process group with the given process ID
        # # Windows does not have a process group. TaskKill will delete a tree of processes
        # #
        # # @param pid   the process ID to kill
        # # @return None
        # def kill(self, pid):
        #     # TODO move to utils_ps; may or may not work on Windows
        #     cmd = ['TASKKILL', '/F', '/T', '/PID', str(pid)]
        #     proc = subprocess.Popen(cmd,  # pylint: disable=consider-using-with
        #                             bufsize=0,
        #                             universal_newlines=True,
        #                             stdin=None,
        #                             stdout=subprocess.PIPE,
        #                             stderr=subprocess.STDOUT)
        #     svc.log.output('OsSpecific:kill', proc.stdout.readlines())

        # # -------------------
        # ## splits a command line argument into a list of multiple arguments, used in CLI argparse
        # #
        # # @param arg  the command line argument to parse
        # # @return the parsed command line as required for the platform
        # def split_args(self, arg):
        #     re_cmd_lex = r'''"((?:""|\\["\\]|[^"])*)"?()|(\\\\(?=\\*")|\\")|(&&?|\|\|?|\d?>|[<])|([^\s"&|<>]+)|(\s+)|(.)'''
        #
        #     args = []
        #     accu = None  # collects pieces of one arg
        #     for quotes, qss, esc, pipe, word, white, fail in re.findall(re_cmd_lex, arg):
        #         if word:
        #             pass  # most frequent
        #         elif esc:
        #             word = esc[1]
        #         elif white or pipe:
        #             if accu is not None:
        #                 args.append(accu)
        #             if pipe:
        #                 args.append(pipe)
        #             accu = None
        #             continue
        #         elif fail:
        #             raise ValueError('invalid or incomplete shell string')
        #         elif quotes:
        #             word = quotes.replace('\\"', '"').replace('\\\\', '\\')
        #             word = word.replace('""', '"')
        #         else:
        #             word = qss  # may be even empty; must be last
        #
        #         accu = (accu or '') + word
        #
        #     if accu is not None:
        #         args.append(accu)
        #
        #     return args

    # -------------------
    ## implements the Mac specific commands
    class Mac:
        # -------------------
        ## constructor
        def __init__(self):
            pass

        # -------------------
        ## holds OS information
        #
        # @return string indicating OS info
        def os_version(self):
            return f'macOS {platform.mac_ver()[0]}'

        # -------------------
        ## simple OS name
        #
        # @return string indicating OS
        def os_name(self):
            return 'macos'

        # # -------------------
        # ## parses the command line into an array of tokens as required by the shell
        # #
        # # @param cmdline the command line to parse
        # # @return an array of elements from the given cmdline
        # def parse_command_line(self, cmdline):
        #     return shlex.split(cmdline)

        # # -------------------
        # ## returns a preexec_fn for Popen calls
        # #
        # # @return the sid required by Popen calls
        # def preexec_fn(self):
        #     return os.setsid

        # # -------------------
        # ## returns the package manager
        # #
        # # @return Brew package manager
        # def package_manager(self):
        #     return 'brew'

        # # -------------------
        # ## kill the process group with the given process ID
        # #
        # # @param pid   the process ID to kill
        # # @return None
        # def kill(self, pid):
        #     os.killpg(os.getpgid(pid), signal.SIGTERM)

        # # -------------------
        # ## splits a command line argument into a list of multiple arguments, used in CLI argparse
        # #
        # # @param arg  the command line argument to parse
        # # @return the parsed command line as required for the platform
        # def split_args(self, arg):
        #     return shlex.split(arg)

    # -------------------
    ## implements the Ubuntu specific commands
    class Ubuntu:
        # -------------------
        ## constructor
        def __init__(self):
            pass

        # -------------------
        ## simple OS name
        #
        # @return string indicating OS
        def os_name(self):
            return 'ubuntu'

        # -------------------
        ## holds OS information
        #
        # @return string indicating OS info
        def os_version(self):
            proc = subprocess.Popen(["lsb_release", "-a"],  # pylint: disable=consider-using-with
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT
                                    )
            (out, _) = proc.communicate()
            out = out.decode('utf-8')

            version = 'notset'
            codename = 'notset'
            for line in out.split('\n'):
                # print(f'line: "{line}"')
                args = line.split('\t')
                if args[0] == 'Release:':
                    version = args[1]
                elif args[0] == 'Codename:':
                    codename = args[1]
            return f'Ubuntu {version} {codename}'

        # # -------------------
        # ## parses the command line into an array of tokens as required by the shell
        # #
        # # @param cmdline the command line to parse
        # # @return an array of elements from the given cmdline
        # def parse_command_line(self, cmdline):
        #     return shlex.split(cmdline)

        # # -------------------
        # ## returns a preexec_fn for Popen calls
        # # not used in Ubuntu implementation
        # #
        # # @return None
        # def preexec_fn(self):
        #     return None

        # # -------------------
        # ## returns the package manager
        # #
        # # @return APT package manager
        # def package_manager(self):
        #     return 'apt'

        # # -------------------
        # ## kill the process group with the given process ID
        # #
        # # @param pid   the process ID to kill
        # # @return None
        # def kill(self, pid):
        #     os.killpg(os.getpgid(pid), signal.SIGTERM)

        # # -------------------
        # ## splits a command line argument into a list of multiple arguments, used in CLI argparse
        # #
        # # @param arg  the command line argument to parse
        # # @return the parsed command line as required for the platform
        # def split_args(self, arg):
        #     return shlex.split(arg)

    # -------------------
    ## implements the Raspberry Pi specific commands
    class RaspberryPi:
        # -------------------
        ## constructor
        def __init__(self):
            pass

        # -------------------
        ## report current OS
        #
        # @return string indicating OS
        def os_name(self):
            return 'rpi'

        # -------------------
        ## holds OS information
        #
        # @return string indicating OS info
        def os_version(self):
            # TODO confirm this is correct
            return f'RPI {platform.system()} {platform.release()}'

        # # -------------------
        # ## parses the command line into an array of tokens as required by the shell
        # #
        # # @param cmdline the command line to parse
        # # @return an array of elements from the given cmdline
        # def parse_command_line(self, cmdline):
        #     return shlex.split(cmdline)

        # # -------------------
        # ## returns a preexec_fn for Popen calls
        # # not used in Ubuntu implementation
        # #
        # # @return None
        # def preexec_fn(self):
        #     return None

        # # -------------------
        # ## returns the package manager
        # #
        # # @return APT package manager
        # def package_manager(self):
        #     return 'apt'

        # # -------------------
        # ## kill the process group with the given process ID
        # #
        # # @param pid   the process ID to kill
        # # @return None
        # def kill(self, pid):
        #     os.killpg(os.getpgid(pid), signal.SIGTERM)

        # # -------------------
        # ## splits a command line argument into a list of multiple arguments, used in CLI argparse
        # #
        # # @param arg  the command line argument to parse
        # # @return the parsed command line as required for the platform
        # def split_args(self, arg):
        #     return shlex.split(arg)

    # -------------------
    ## initialize
    #
    # selects the current platform and sets impl to it
    # @return None
    @classmethod
    def init(cls):
        if os.path.isfile('/sys/firmware/devicetree/base/model'):
            cls.impl = OsSpecific.RaspberryPi()
        elif sys.platform == 'win32':
            ## holds the implementation class
            cls.impl = OsSpecific.Win()
        elif sys.platform == 'darwin':
            cls.impl = OsSpecific.Mac()
        elif sys.platform == 'linux':
            cls.impl = OsSpecific.Ubuntu()
        else:
            svc.log.bug(f'unrecognized OS: "{sys.platform}"')
            sys.exit(1)

        ## holds the simple OS name: win, ubuntu, macos, rpi
        cls.os_name = cls.impl.os_name()
        ## holds OS information e.g. "ubuntu 20.04 focal"
        cls.os_version = cls.impl.os_version()

        ## holds python version e.g. "3.10"
        cls.python_version = f'Python {sys.version_info.major}.{sys.version_info.minor}'

    # -------------------
    ## get current userid
    #
    # @return userid
    @classmethod
    def userid(cls):
        return getpass.getuser()

    # -------------------
    ## get current hostname
    #
    # @return hostname
    @classmethod
    def hostname(cls):
        return platform.uname().node

    # # -------------------
    # ## parses a command line to be passed into a process start e.g. Popen
    # #
    # # @param cmdline  the command line to parse
    # # @return the parsed command line as required for the platform
    # @classmethod
    # def parse_command_line(cls, cmdline):
    #     return cls.impl.parse_command_line(cmdline)

    # # -------------------
    # ## returns a preexec_fn for Popen calls
    # #
    # # @return the preexec_fn needed for the platform
    # @classmethod
    # def preexec_fn(cls):
    #     return cls.impl.preexec_fn()

    # # -------------------
    # ## returns the name of the package manager
    # #
    # # @return the package manager for the platform
    # @classmethod
    # def package_manager(cls):
    #     return cls.impl.package_manager()

    # # -------------------
    # ## kill the process tree/group with the given process ID
    # #
    # # @param pid   the process ID to kill
    # # @return None
    # @classmethod
    # def kill(cls, pid):
    #     cls.impl.kill(pid)

    # # -------------------
    # ## splits a command line argument into a list of multiple arguments, used in CLI argparse
    # #
    # # @param arg  the command line argument to parse
    # # @return the parsed command line as required for the platform
    # @classmethod
    # def split_args(cls, arg):
    #     return cls.impl.split_args(arg)



from .svc import svc


# --------------------
## perform the do_makefile operations
class DoMakefile:
    # --------------------
    ## do_build mainline.
    #
    # @param target      cmake target to build e.g. ut
    # @return None
    def run(self, target):
        svc.log.highlight(f'{svc.gbl.tag}: starting target:{target}...')
        self._run_make(target)
        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: target {target} rc={svc.gbl.rc}')

    # --------------------
    ## run a makefile target
    #
    # @param target      the make target to run e.g. ut
    # @return None
    def _run_make(self, target):
        cmd = f'make {target}'
        svc.utils_ps.run_process(cmd)



import os

from .do_cpip import DoCpip
from .svc import svc


# --------------------
## perform the do_lint operation
class DoBuild:
    # --------------------
    ## do_build mainline.
    #
    # @param tech        technology: cpp, arduino
    # @param build_type  build type: debug or release
    # @param target      cmake target to build e.g. ut
    # @return None
    def run(self, tech, build_type, target):
        if not build_type:
            build_type = 'debug'

        if build_type not in ['debug', 'release']:
            svc.gbl.rc += 1
            svc.log.err(f'unknown build_type:{build_type}, use one of: debug, release')

        if not tech:
            # default to xplat.cfg value
            tech = svc.cfg.mod_tech
        if not tech:
            # not set, the default to cpp
            tech = 'cpp'

        if not target:
            target = svc.cfg.mod_name

        svc.log.highlight(f'{svc.gbl.tag}: starting tech:{tech} target:{target}...')

        # generate CPIP files as needed
        cpip = DoCpip()
        cpip.gen()

        if tech in ['cpp']:
            self._run_cpp(build_type, target)
        elif tech in ['arduino']:
            self._run_arduino(build_type, target)
        else:
            svc.gbl.rc += 1
            svc.log.err(f'unknown tech:{tech}, use one of: python, cpp, arduino')

        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: build {build_type} rc={svc.gbl.rc}')

    # --------------------
    ## build a cpp app or library
    #
    # @param build_type  build type: debug or release
    # @param target      the cmake target to build e.g. ut
    # @return None
    def _run_cpp(self, build_type, target):
        build_dir = f'cmake-build-{build_type}'
        if not os.path.isdir(build_dir):
            svc.utils_fs.safe_create_dir(build_dir)
            if build_type == 'debug':
                btype = 'Debug'
            else:
                btype = 'Release'
            svc.log.line(f'{svc.gbl.tag}: load project for {build_type}')
            cmd = f'cmake -DCMAKE_BUILD_TYPE={btype} -DCMAKE_MAKE_PROGRAM=ninja -G Ninja -S . -B {build_dir}'
            svc.utils_ps.run_process(cmd, use_raw_log=False)
            svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: cmake load proj. {build_type} rc={svc.gbl.rc}')

        svc.log.line(f'{svc.gbl.tag}: update makefile')
        cmd = f'cmake -DCMAKE_EXPORT_COMPILE_COMMANDS=ON {build_dir}/'
        svc.utils_ps.run_process(cmd, use_raw_log=False)
        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: cmake makefile {build_type} rc={svc.gbl.rc}')

        if svc.gbl.rc != 0:
            return

        svc.log.line(f'{svc.gbl.tag}: build {build_type} target:{target}')
        cmd = f'cmake --build {build_dir} --target {target} -j 6'
        svc.utils_ps.run_process(cmd, use_raw_log=False)
        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: cmake build {build_type} rc={svc.gbl.rc}')

    # --------------------
    ## build an arduino binary image
    #
    # @param build_type  build type: debug or release
    # @param target      the cmake target to build
    # @return None
    def _run_arduino(self, build_type, target):
        svc.log.line(f'{svc.gbl.tag}: build {build_type} target:{target}')
        if target == 'ut':
            make_parm = 'ut_build'
        else:
            make_parm = f'build-{build_type}'
        cmd = f'make {make_parm}'
        svc.utils_ps.run_process(cmd, use_raw_log=False)
        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: arduino {make_parm} rc={svc.gbl.rc}')



import os
import re

from .svc import svc


# --------------------
## perform the do_clean operation
class DoPublish:
    # --------------------
    ## do_publish mainline.
    #
    # @return None
    def run(self):
        # script also prints this; comment out for now
        # log.highlight(f'{svc.glbl.tag} starting...')
        svc.gbl.rc = 0

        if svc.cfg.mod_name in svc.cfg.do_publish.disallow:
            svc.log.warn(f'cannot publish; disallowed by: {svc.cfg.do_publish.disallow}')
            svc.gbl.rc = 1
            return

        if not svc.cfg.is_module:
            svc.log.warn('do_publish is only functional for modules')

        self._gen_setup_py()
        self._gen_manifest_in()

    # --------------------
    ## generate setup.py file
    #
    # @return None
    def _gen_setup_py(self):
        svc.utils_fs.safe_delete_file('setup.py')

        # generate the setup.py file from setup_template.py
        src = os.path.join(svc.utils_fs.root_dir, 'tools', 'setup_template.py')
        dst = os.path.join(svc.utils_fs.root_dir, 'setup.py')
        with open(dst, 'w', encoding='utf-8', newline='\n') as fp_dst:
            with open(src, 'r', encoding='utf-8') as fp_src:
                for line in fp_src:
                    while True:
                        m = re.search(r'\$\$([^$]+)\$\$', line)
                        if not m:
                            # uncomment to debug
                            # log.dbg(f'line: {line.rstrip()}')
                            fp_dst.write(line)
                            break

                        vrbl = m.group(1)
                        if vrbl == 'root_dir':
                            val = str(svc.utils_fs.root_dir)
                        else:
                            val = getattr(svc.cfg, vrbl)
                        # uncomment to debug
                        # log.dbg(f'grp:{m.group(1)} line:{line.strip()}')
                        line = line.replace(f'$${vrbl}$$', val)

    # --------------------
    ## generate manifest file
    #
    # @return None
    def _gen_manifest_in(self):
        fname = 'MANIFEST.in'
        svc.utils_fs.safe_delete_file(fname)
        dst = os.path.join(svc.utils_fs.root_dir, fname)
        with open(dst, 'w', encoding='utf-8', newline='\n') as fp:
            # graft module_name
            # global-exclude *.py[cod]
            fp.write(f'graft {svc.cfg.mod_dir_name}\n')
            fp.write('global-exclude *.py[cod]\n')



import json
import os

from .svc import svc


# --------------------
## perform the do_post_ver operation
class DoPostVer:
    # --------------------
    ## do_post_ver mainline.
    #
    # @return None
    def run(self):
        svc.log.highlight(f'{svc.gbl.tag}: starting...')

        slug = svc.cfg.do_post_ver.slug
        if slug.lower() in ['none', 'n/a', 'na']:
            svc.log.warn(f'{svc.gbl.tag}: not a web project, skipping (slug: "{slug}")')
            return

        # ensure web-ver is created
        svc.utils_fs.safe_global_dir(svc.cfg.do_post_ver.dst_dir)
        dst_dir = os.path.expanduser(svc.cfg.do_post_ver.dst_dir)

        self._gen_info_json(slug, dst_dir)
        self._save_verification_files(slug, dst_dir)

        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: verification files post, rc={svc.gbl.rc}')
        svc.gbl.overallrc += svc.gbl.rc

    # --------------------
    ## create an info.json file to contain various information
    #
    # @param slug     the web page slug
    # @param dst_dir  the destination directory to save the json file
    # @return None
    def _gen_info_json(self, slug, dst_dir):
        info = {
            'slug': slug,
            'version': svc.cfg.version,
            'pypi_url': svc.cfg.do_post_ver.pypi_url,
            'repo_url': svc.cfg.do_post_ver.repo_url,
        }

        info_path = os.path.join(dst_dir, f'{slug}-info.json')
        with open(info_path, 'w', encoding='utf-8', newline='\n') as fp:
            json.dump(info, fp)
        svc.log.line(f'{svc.gbl.tag}: gen info: {info_path}')

    # --------------------
    ## copy verification html files to destination directory
    # assume: skeleton HTML files were generated by medver-pytest
    #
    # @param slug     the web page slug
    # @param dst_dir  the destination directory to save the json file
    # @return None
    def _save_verification_files(self, slug, dst_dir):
        src_dir = os.path.expanduser(svc.cfg.do_post_ver.src_dir)
        self._save_verification_file(slug, dst_dir, 'summary', src_dir)
        self._save_verification_file(slug, dst_dir, 'trace', src_dir)
        self._save_verification_file(slug, dst_dir, 'report', src_dir, src_fname='test_report')

        # TODO any other files go here

    # --------------------
    ## save verification file to destination directory
    # assume: skeleton HTML files were generated by medver-pytest
    #
    # @param slug       the web page slug
    # @param dst_dir    the destination directory to save the html file
    # @param fname      the file to save (no extension)
    # @param src_dir    the source directory (usually out/ver)
    # @param src_fname  the source file name, if different from fname
    # @return None
    def _save_verification_file(self, slug, dst_dir, fname, src_dir, src_fname=None):
        if src_fname is None:
            src_path = os.path.join(src_dir, f'{fname}-min.html')
        else:
            src_path = os.path.join(src_dir, f'{src_fname}-min.html')

        if not os.path.isfile(src_path):
            svc.log.warn(f'{svc.gbl.tag}: src not found, skipping: {src_path}')
            return

        # note extension changed to html
        dst_fname = f'{slug}-{fname}.html'
        svc.utils_fs.safe_copy_file(src_path, dst_dir, dst_fname)
        svc.log.line(f'{svc.gbl.tag}: saved: {os.path.join(dst_dir, dst_fname)}')



import logging
import logging.handlers

import sys


class RsyslogLogger:
    def __init__(self):
        # Gets or creates a logger
        self._logger = logging.getLogger(__name__)

        # set log level
        self._logger.setLevel(logging.DEBUG)

        # formatter = logging.Formatter('%(asctime)s : %(levelname)s : %(name)s : %(message)s')

        sh = logging.handlers.SysLogHandler(address='/dev/log')
        # sh.setFormatter(formatter)
        self._logger.addHandler(sh)

        stdout_handler = logging.StreamHandler(sys.stdout)
        # handler.setFormatter(formatter)
        self._logger.addHandler(stdout_handler)

    def term(self):
        pass

    def debug(self, msg=''):
        self._logger.debug(msg)


def create_logger(app_name, log_level=logging.DEBUG, stdout=True, syslog=False, file=False):
    # create logger
    logger = logging.getLogger(app_name)
    logger.setLevel(log_level)

    # set log format to handlers
    formatter = logging.Formatter('%(name)s - %(asctime)s - %(levelname)s - %(message)s')

    if file:
        # create file logger handler
        fh = logging.FileHandler('my-sample-app.log')
        fh.setLevel(log_level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    if syslog:
        # create syslog logger handler
        sh = logging.handlers.SysLogHandler(address='/dev/log')
        sh.setLevel(log_level)
        sf = logging.Formatter('%(name)s: %(message)s')
        sh.setFormatter(sf)
        logger.addHandler(sh)

    if stdout:
        # create stream logger handler
        ch = logging.StreamHandler()
        ch.setLevel(log_level)
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    return logger



import queue
import sys
import threading
import time
import traceback


# --------------------
## holds logging functions that replace common python logger functions
class FalconLogger:

    # --------------------
    ## constructor
    #
    # @param path        None for stdout, or full path to the logger file
    # @param max_entries (optional) maximum number of entries before a flush is done; default 10
    # @param loop_delay  (optional) time between checking queue; default 0.250 seconds
    def __init__(self, path=None, max_entries=10, loop_delay=0.250):
        ## the full path to the log file
        self._path = path

        ## the maximum entries to hold in the queue before saving to the file
        self._max_entries = max_entries
        if self._max_entries <= 0:
            raise Exception('max_entries must be greater than 0')  # pylint: disable=broad-exception-raised

        ## the delay between checking the queue for entries to save
        self._loop_delay = loop_delay
        if self._loop_delay < 0.100:
            raise Exception('loop_delay must be >= 0.100 seconds')  # pylint: disable=broad-exception-raised

        # print every second even if less than max_entries are in the queue
        ## the maximum number of loops before the queue is emptied
        self._max_count = int(round(1 / self._loop_delay, 1))

        ## the queue
        self._queue = queue.Queue()
        ## the file pointer
        if self._path is None:
            self._fp = sys.stdout
        else:
            self._fp = open(self._path, 'w', encoding='UTF-8')  # pylint: disable=consider-using-with

        ## flag to the thread to end the loop
        self._finished = False
        ## the thread pointer
        self._thread = threading.Thread(target=self._runner)
        self._thread.daemon = True
        self._thread.start()
        # wait for thread to start
        time.sleep(0.1)

    # --------------------
    ## log a debug line
    #
    # @param msg the line to print; default empty
    # @return None
    def debug(self, msg=''):
        self._queue.put(msg)

    # --------------------
    ## log an info line
    #
    # @param msg the line to print; default empty
    # @return None
    def info(self, msg=''):
        self._queue.put(msg)

    # --------------------
    ## log a warning line
    #
    # @param msg the line to print; default empty
    # @return None
    def warning(self, msg=''):
        self._queue.put(msg)

    # --------------------
    ## log an error line
    #
    # @param msg the line to print; default empty
    # @return None
    def error(self, msg=''):
        self._queue.put(msg)

    # --------------------
    ## log a critical line
    #
    # @param msg the line to print; default empty
    # @return None
    def critical(self, msg=''):
        self._queue.put(msg)

    # --------------------
    ## log an exception
    #
    # @param excp the exception to print
    # @return None
    def exception(self, excp):
        for line in traceback.format_exception(excp):
            self._queue.put(line)

    # --------------------
    ## terminate
    # stop the thread, save any remaining line in the internal queue
    #
    # @return None
    def term(self):
        self._finished = True
        if self._thread.is_alive():  # pragma: no cover
            self._thread.join(5)

    # --------------------
    ## do a save at this point
    #
    # @return None
    def save(self):
        self._save()

    # --------------------
    ## the thread runner
    # wakes periodically to check if the queue has max_entries or more in it
    # if so, the lines are written to the file
    # if not, it sleeps
    #
    # @return None
    def _runner(self):
        count = 0
        while not self._finished:
            if count < self._max_count and self._queue.qsize() < self._max_entries:
                count += 1
                time.sleep(self._loop_delay)
                continue

            count = 0
            self._save()

        self._save()
        if self._path:
            self._fp.close()
            self._fp = None

    # --------------------
    ## save any entries in the queue to the file
    #
    # @return None
    def _save(self):
        if self._fp is None:
            self._finished = True
            return

        count = self._queue.qsize()
        while count > 0:
            msg = self._queue.get_nowait()
            # TODO print(msg)
            self._fp.write(msg)
            self._fp.write('\n')
            count -= 1
        self._fp.flush()



import os

from .svc import svc


# --------------------
## perform gen constants operations
class GenConstants:
    # --------------------
    ## generate constant file
    #
    # @param subdir  (optional) subdirectory to place the constants files
    # @return None
    def run(self, subdir):
        if svc.cfg.mod_tech in ['', 'python']:
            self.gen_constants(subdir)
        else:
            svc.log.warn(f'gen_constants: does not work for tech: {svc.cfg.mod_tech}')
            svc.gbl.rc = 1

    # --------------------
    ## generate the constants_version.py file in the module/app directory
    #
    # @param subdir the subdirectory in the source root directory
    # @return None
    def gen_constants(self, subdir):
        if subdir:
            filedir = os.path.join('lib', subdir)
        else:
            filedir = os.path.join('lib')

        if svc.cfg.is_module:
            mod_dir = os.path.join(svc.utils_fs.root_dir, svc.cfg.mod_dir_name)
            if not os.path.isdir(mod_dir):
                svc.log.warn(f'gen_constants: dir does not exist: {mod_dir}')
                svc.gbl.rc = 1
                return

            path = os.path.join(str(mod_dir), filedir)
        else:
            path = os.path.join(svc.utils_fs.root_dir, filedir)

        path = os.path.join(path, 'constants_version.py')
        # svc.log.dbg(f'path: {path}')

        # generate constants file
        with open(path, 'w', encoding='utf-8', newline='\n') as fp:
            fp.write('from dataclasses import dataclass\n')
            fp.write('\n')
            fp.write('\n')
            fp.write('# --------------------\n')
            fp.write('## holds constants\n')
            fp.write('@dataclass\n')
            fp.write('class ConstantsVersion:\n')
            fp.write('    ## current App version\n')
            fp.write(f'    version = \'{svc.cfg.version}\'\n')

        if os.path.isfile(path):
            svc.gbl.rc = 0
        else:
            svc.gbl.rc = 1

        msg = f'gen: constants_version.py rc={svc.gbl.rc}'
        if not svc.gbl.verbose:
            svc.log.line(msg)
        elif svc.gbl.rc == 0:
            svc.log.ok(msg)
        else:
            svc.log.warn(msg)
        svc.gbl.overallrc += svc.gbl.rc



from .svc import svc


# --------------------
## perform the do_coverage operations for Makefile based projects
class DoMakefileCoverage:
    # --------------------
    ## do_makefile_coverage mainline.
    #
    # @param action  the action to take: reset, gen
    # @param target  the makefile target name
    # @return None
    def run(self, action, target):
        svc.log.highlight(f'{svc.gbl.tag}: starting {action} for target: {target}...')

        if action == 'reset':
            self._cpp_reset_coverage(target)
        elif action == 'gen':
            self._cpp_gen_coverage(target)
        else:
            svc.gbl.rc += 1
            svc.log.err(f'unknown action:{action}, use one of: reset, gen')

        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: action:{action} rc={svc.gbl.rc}')

    # --------------------
    ## reset gcov counters by deleting debug/*.gcda files
    #
    # @param target  the target to generate the coverage for
    # @return None
    def _cpp_reset_coverage(self, target):
        cmd = f'make {target}-cov-reset'
        svc.utils_ps.run_process(cmd)

    # --------------------
    ## generate coverage using gcovr.
    # see https://gcovr.com/en/4.2/guide.html
    # * Note: gcovr --gcov-executable gcov-8
    # * Note: gcovr --gcov-executable "llvm-cov gcov" # For CLANG
    #
    # @param target  the target to generate the coverage for
    # @return None
    def _cpp_gen_coverage(self, target):
        cmd = f'make {target}-cov'
        svc.utils_ps.run_process(cmd)



import os
import re
import sys

from .os_specific import OsSpecific
from .svc import svc


# --------------------
## generates the build_info information
class GenBuildInfo:
    # --------------------
    ## constructor
    def __init__(self):
        ## the file pointer for the build_info file
        self._fp = None
        ## the path to the build_info file
        self._path = None
        ## the overall exit return code
        self._exitrc = 0

    # --------------------
    ## generate build_info file
    #
    # @param subdir  (optional) subdirectory to place the constants files
    # @return None
    def run(self, subdir):
        if svc.cfg.mod_tech in ['', 'python']:
            self.gen_build_info(subdir, 'build_info.py')
        elif svc.cfg.mod_tech in ['cpp', 'arduino']:
            self.gen_build_info(subdir, 'build_info.txt')
        else:
            svc.log.warn(f'gen_build_info: does not work for tech: {svc.cfg.mod_tech}')
            svc.gbl.rc = 1

    # --------------------
    ## generate the build_info.py file in the module/app directory
    #
    # @param subdir   the subdirectory in the source root directory
    # @param fname    the build_info filename to use
    # @return None
    def gen_build_info(self, subdir, fname):
        if subdir:
            filedir = os.path.join('lib', subdir)
        else:
            filedir = os.path.join('lib')

        if svc.cfg.is_module:
            mod_dir = os.path.join(svc.utils_fs.root_dir, svc.cfg.mod_dir_name)
            if not os.path.isdir(mod_dir):
                svc.log.warn(f'gen_build_info: dir does not exist: {mod_dir}')
                svc.gbl.rc = 1
                return

            path = os.path.join(str(mod_dir), filedir)
        else:
            path = os.path.join(svc.utils_fs.root_dir, filedir)

        path = os.path.join(path, fname)
        # svc.log.dbg(f'path: {path}')

        # generate file
        self.init(path)
        svc.gbl.rc = self.gen(svc.cfg.version)
        self.term()

        if os.path.isfile(path):
            svc.gbl.rc = 0
        else:
            svc.gbl.rc = 1

        msg = f'gen: {fname} rc={svc.gbl.rc}'
        if not svc.gbl.verbose:
            svc.log.line(msg)
        elif svc.gbl.rc == 0:
            svc.log.ok(msg)
        else:
            svc.log.warn(msg)
        svc.gbl.overallrc += svc.gbl.rc

    # --------------------
    ## initialize
    #
    # @param path     the path to the build_info file
    # @return None
    def init(self, path):
        self._path = path

        self._fp = open(self._path, 'w', encoding='utf-8', newline='\n')  # pylint: disable=consider-using-with

    # --------------------
    ## terminate
    #
    # @return None
    def term(self):
        if self._fp is not None:
            self._fp.close()

    # --------------------
    ## generate all build info
    #
    # @param version   the module/app version
    # @return error return code
    def gen(self, version):
        self._gen_init(version)
        self._gen_git_sha()
        self._gen_git_branch()
        self._gen_uncommitted_changes()
        self._gen_unpushed_commits()
        if svc.cfg.mod_tech != 'python':
            self._writeln('    === */')

        return self._exitrc

    # --------------------
    ## generate common build values
    #
    # @param version   the module/app version
    # @return None
    def _gen_init(self, version):
        if svc.cfg.mod_tech == 'python':
            self._writeln('class BuildInfo:  # pylint: disable=too-few-public-methods')
        else:
            self._writeln('/* === Build Info:')

        m = re.search(r'^(\d+\.\d+\.\d+) ', sys.version)
        self._set('python version', 'python_version', m.group(1))
        self._set('OS name', 'os_name', f'{os.name}:{OsSpecific.os_name}')

        # ensure file is created & flushed here so the import works cleanly
        self._fp.flush()

        self._set('version', 'version', version)

    # --------------------
    ## set the value in BuildInfo object
    #
    # @param tag    tag for logging
    # @param name   the name of the variable
    # @param val    the value of the variable
    # @return None
    def _set(self, tag, name, val):
        if svc.gbl.verbose:
            svc.log.ok(f'{tag: <25}: {val}')
        self._writeln(f'    {name} = \'{val}\'')

    # --------------------
    ## set a list of value in BuildInfo object
    #
    # @param name   the name of the list variable
    # @param items  the values of the list variable
    # @return None
    def _setlist(self, name, items):
        self._writeln(f'    {name} = [')
        for item in items:
            self._writeln(f'        \'{item}\',')
        self._writeln('    ]')

    # --------------------
    ## write a line to the build_info file
    # ensures it is terminated with a linefeed
    #
    # @param line  the line to write
    # @return None
    def _writeln(self, line):
        self._fp.write(f'{line}\n')

    # --------------------
    ## gen the current git SHA for the latest commit
    #
    # @return None
    def _gen_git_sha(self):
        tag = 'git SHA'
        rc, out = svc.utils_ps.run_cmd('git rev-parse --verify HEAD')
        self._exitrc += rc

        if rc != 0:
            svc.log.err(f'{tag: <25}: cmd failed: rc={rc}')
            return

        self._set(tag, 'git_sha', out)

    # --------------------
    ## gen the current branch name
    #
    # @return None
    def _gen_git_branch(self):
        tag = 'git branch'

        rc, out = svc.utils_ps.run_cmd('git rev-parse --abbrev-ref HEAD')
        self._exitrc += rc

        if rc != 0:
            svc.log.err(f'{tag: <25}: cmd failed: rc={rc}')
            return

        self._set(tag, 'git_branch', out)

    # --------------------
    ## show any uncommitted changes
    #
    # @return None
    def _gen_uncommitted_changes(self):
        tag = 'git uncommitted changes'
        rc, out = svc.utils_ps.run_cmd('git status -s')
        self._exitrc += rc

        if rc != 0:
            svc.log.err(f'{tag: <25}: cmd failed: rc={rc}')
            return

        uncommitted = []
        count = self._get_failed_lines(tag, out, uncommitted, 'has uncommitted changes')
        self._exitrc += count

        self._setlist('git_uncommitted', uncommitted)

    # --------------------
    ## show any unpushed commits
    #
    # @return None
    def _gen_unpushed_commits(self):
        tag = 'git unpushed commits'
        rc, out = svc.utils_ps.run_cmd('git cherry -v')
        self._exitrc += rc

        if rc != 0:
            svc.log.err(f'{tag: <25}: cmd failed: rc={rc}')
            return

        unpushed = []
        count = self._get_failed_lines(tag, out, unpushed, 'has unpushed changes')
        self._exitrc += count

        self._setlist('git_unpushed', unpushed)

    # --------------------
    ## gather and report failed lines in the given result lines
    #
    # @param tag           the logging tag
    # @param out           the output lines
    # @param failed_lines  the list to append any failed lines
    # @param suffix        the logging suffix if a warning occurs
    # @return the number of failed lines
    def _get_failed_lines(self, tag, out, failed_lines, suffix):
        header = False
        count = 0
        for line in out.split('\n'):
            if line != '':
                count += 1
                if not header:
                    if svc.gbl.verbose:
                        svc.log.warn(f'{tag: <25}:')
                    header = True
                failed_lines.append(line)
                if svc.gbl.verbose:
                    svc.log.warn(f'    {line}')

        if svc.gbl.verbose:
            if count == 0:
                svc.log.ok(f'{tag: <25}: none')
            else:
                svc.log.warn(f'{tag: <25}: {suffix}')
        return count



# -------------------
def loader_init():
    import os
    import inspect

    # get caller's name
    caller = 'unknown'
    for frame in inspect.stack()[1:]:
        if frame.filename[0] != '<':
            caller = os.path.basename(frame.filename)
            break

    # ensure using local or venv as requestted
    if 'VER_LOCAL' in os.environ:
        # assumes that PYTHONPATH has prepended '.'
        print(f'==== using local: {caller}')
    else:
        print(f'==== using venv: {caller}')
        import sys

        # override the PYTHONPATH setting
        sys.path.insert(0, os.path.join('venv', 'lib', 'site-packages'))



import os
import sys

# these must NOT require external imports otherwise will fail when venv doesn't exist
from .cfg import Cfg
from .do_clean import DoClean
from .gen_files import GenFiles
from .svc import svc
from .utils_common import UtilsCommon
from .utils_fs import UtilsFs
from .utils_logger import UtilsLogger
from .utils_ps import UtilsPs


# --------------------
## holds mainline for all utilities
class Utils(UtilsCommon):
    # --------------------
    ## gen bash set_cfg.sh file for use in bash scripts
    #
    # @return None
    def gen_bash_cfg(self):
        self._init('gen_bash_cfg')

        path = os.path.join('tools', 'set_cfg.sh')
        with open(path, 'w', encoding='utf-8', newline='\n') as fp:
            fp.write('#! /usr/bin/env bash\n')

            val = self.get_cfg('mod_name')
            fp.write(f'export cmn_mod_name="{val}"\n')

            val = self.get_cfg('mod_dir_name')
            fp.write(f'export cmn_mod_dir_name="{val}"\n')

            val = str(svc.utils_fs.convert_relative_to_abs(svc.gbl.outdir))
            fp.write(f'export OUT_DIR="{val}"\n')

            val = self.get_cov_opts()
            fp.write(f'export COV_OPTS="{val}"\n')

    # --------------------
    ## create the output directory
    #
    # @return None
    def create_outdir(self):
        self._init('create_out_dir')
        svc.utils_fs.create_outdir()
        # Note: no _term()

    # --------------------
    ## recreate subdirectory in output directory
    #
    # @param subdir the subdirectory to create
    # @return None
    def recreate_outsubdir(self, subdir):
        self._init('recreate_outsubdir')
        svc.utils_fs.clean_out_dir(subdir)
        # Note: no _term()

    # --------------------
    ## check if any python pip modules are out of date
    #
    # @return None
    def pip_outofdate(self):
        self._init('pip_outofdate')
        svc.log.line('out of date modules in tools/requirements.txt')
        # Note: old 'pip3 install --upgrade -r tools/requirements.txt --dry-run --quiet'
        cmd = 'pip3 list --outdated'
        svc.utils_ps.run_process(cmd)
        svc.log.check(svc.gbl.rc == 0, f'out of date rc={svc.gbl.rc}')
        # Note: no _term()

    # --------------------
    ## initialize cfg values and generate version.json
    #
    # @param dogen   generate files
    # @param subdir  (optional) subdir to place files into
    # @return None
    def gen_files(self, dogen='', subdir=''):
        self._set_params(False)
        self._init('gen_files')
        svc.gen_files.all(dogen=dogen, subdir=subdir)
        # Note: no _term()

    # TODO delete after GenConstants and GenBuildInfo merged into GenFiles
    # # --------------------
    # ## initialize generate constants files
    # #
    # # @param subdir  (optional) subdir to place files into
    # # @return None
    # def gen_constants(self, subdir=None):
    #     from .gen_constants import GenConstants
    #     self._set_params(False)
    #     self._init('gen_constants')
    #     impl = GenConstants()
    #     impl.run(subdir)
    #     self._term(False)
    #
    # # --------------------
    # ## initialize generate constants files
    # #
    # # @param subdir  (optional) subdir to place files into
    # # @return None
    # def gen_build_info(self, subdir=None):
    #     from .gen_build_info import GenBuildInfo
    #     self._set_params(False)
    #     self._init('gen_build_info')
    #     impl = GenBuildInfo()
    #     impl.run(subdir)
    #     self._term(False)

    # --------------------
    ## mainline for the do_clean operation.
    # removes all generated files and subdirectories.
    # The root directory should be equivalent to a clean git clone
    #
    # @param subdir  directory where gen_files are placed
    # @param depth   cleaning action depth: full, lite
    # @return None
    def do_clean(self, subdir='', depth=''):
        self._init('do_clean')
        impl = DoClean()
        impl.run(subdir=subdir, depth=depth)
        self._term(True)

    # --------------------
    ## mainline for the do_check operation.
    # checks all versions
    #
    # @return None
    def do_check(self):
        verbose = True
        from .do_check import DoCheck
        self._set_params(verbose)
        self._init('do_check')
        impl = DoCheck()
        impl.run()
        self._term(verbose)

    # --------------------
    ## mainline for the do_doc operation.
    # runs doxygen on all files specified in Doxyfile
    #
    # @param make_pdf  generate the PDF (default) or not (False)
    # @return None
    def do_doc(self, make_pdf=True):
        from .do_doc import DoDoc
        self._set_params(True)
        self._init('do_doc')
        impl = DoDoc()
        impl.run(make_pdf)
        self._term(True)

    # --------------------
    ## mainline for the do_lint operation.
    # runs pylint on all given source files
    #
    # @param dogen  generate files or not
    # @param subdir (optional) subdir to place files into
    # @param tool  for cpp/arduino use clang-tidy or cppcheck
    # @return None
    def do_lint(self, dogen='', subdir='', tool=''):
        from .do_lint import DoLint
        verbose = False
        self._set_params(verbose)
        self._init('do_lint')
        svc.gen_files.all(dogen=dogen, subdir=subdir)
        impl = DoLint()
        impl.run(tool)
        self._term(verbose)

    # --------------------
    # DEPRECATED but still used by swig-c and swig-cpp
    ## mainline for the do_build operation.
    # runs cmake on source files
    #
    # @param verbose     True for additional verbosity
    # @param build_type  build type: debug or release
    # @param tech        technology: cpp or arduino
    # @param target      the cmake target to build e.g. ut
    # @return None
    def do_build(self, verbose, build_type='debug', tech=None, target=None):
        from .do_build import DoBuild
        self._set_params(verbose)
        self._init('do_build')
        impl = DoBuild()
        impl.run(tech, build_type, target)
        self._term(verbose)

    # --------------------
    ## mainline for the do_ver_info operation.
    # generate version info files
    #
    # @return None
    def do_ver_info(self):
        verbose = True
        from .do_ver_info import DoVerInfo
        self._set_params(verbose)
        self._init('do_ver_info')
        impl = DoVerInfo()
        impl.run()
        self._term(verbose)

    # --------------------
    ## mainline for the do_publish operation.
    # post the verification data
    #
    # @return None
    def do_post_ver(self):
        verbose = True
        from .do_post_ver import DoPostVer
        self._set_params(verbose)
        self._init('do_post_ver')
        impl = DoPostVer()
        impl.run()
        self._term(verbose)

    # --------------------
    ## mainline for the do_publish operation.
    # generate the setup.py file
    #
    # @param dogen  generate files or not
    # @param subdir (optional) subdir to place files into
    # @return None
    def do_publish(self, dogen='', subdir=''):
        verbose = False
        from .do_publish import DoPublish
        self._set_params(verbose)
        self._init('do_publish')
        svc.gen_files.all(dogen=dogen, subdir=subdir)
        impl = DoPublish()
        impl.run()
        # script also reports overallrc; set verbose False to prevent
        self._term(False)

    # === do_xx: UT related

    # note python does not have a do_ut() see below for C++/Arduino

    # --------------------
    ## mainline for the do_coverage operation.
    #
    # @param action      reset or gen
    # @return None
    def do_coverage(self, action=None):
        verbose = True
        from .do_coverage import DoCoverage
        self._set_params(verbose)
        self._init('do_coverage')
        impl = DoCoverage()
        impl.run(action)
        self._term(False)

    # === C++/Arduino related

    # --------------------
    ## mainline for the do_makefile operation.
    #
    # @param target      the makefile target to build e.g. all
    # @return None
    def do_makefile(self, target='all'):
        verbose = False
        from .do_makefile import DoMakefile
        self._set_params(verbose)
        self._init('do_makefile')
        impl = DoMakefile()
        impl.run(target)
        self._term(verbose)

    # === C++/Arduino UT related

    # --------------------
    ## mainline for the do_makefile_coverage operations.
    #
    # @param action      reset or gen
    # @param target      the target to generate coverage for
    # @return None
    def do_makefile_coverage(self, action, target):
        verbose = False
        from .do_makefile_coverage import DoMakefileCoverage
        self._set_params(verbose)
        self._init('do_makefile_coverage')
        impl = DoMakefileCoverage()
        impl.run(action, target)
        self._term(False)

    # === CPIP related

    # --------------------
    ## mainline for the do_cpip operation.
    #
    # @param action      publish or pull
    # @return None
    def do_cpip(self, action):
        verbose = True
        from .do_cpip import DoCpip
        self._set_params(verbose)
        self._init('do_cpip')
        impl = DoCpip()
        impl.run(action)
        self._term(verbose)

    # --------------------
    ## get cpip pkg info
    #
    # @param pkg_name   the package to get info about
    # @return None
    def cpip_get(self, pkg_name):
        from .do_cpip import DoCpip
        self._set_params(False)
        self._init('do_cpip')
        impl = DoCpip()
        return impl.get(pkg_name)

    # === common utility functions

    # TODO move to utils_common or to utils_fs
    # --------------------
    ## takes the given path and converts it a path relative to home, e.g,
    # * ~/out    => ~/out
    # * ./out    => ~/projects/proj1/out
    # * out      => ~/projects/proj1/out
    #
    # @param path the path to convert
    # @return the converted path
    def convert_relative_to_home(self, path):
        path1 = os.path.expanduser(path)
        path2 = os.path.abspath(path1)
        homedir = os.path.expanduser('~')
        path3 = path2.replace(homedir, '~')
        return path3

    # === support functions related

    # --------------------
    ## set the class parameters used to run mainline functions
    #
    # @param verbose       whether the function should run with verbose logging or not
    # @return None
    def _set_params(self, verbose):
        svc.gbl.verbose = verbose

    # --------------------
    ## initialization.
    # sets logging tag and overallrc to 0
    #
    # @param tag  the logging tag to use
    # @return None
    def _init(self, tag):
        svc.gbl.tag = tag
        svc.gbl.rc = 0
        svc.gbl.overallrc = 0

        # must be before Cfg, etc.
        svc.log = UtilsLogger

        svc.cfg = Cfg
        svc.utils_fs = UtilsFs()
        svc.utils_ps = UtilsPs()
        svc.gen_files = GenFiles()

        # load cfg if necessary
        svc.cfg.load(svc.utils_fs.root_dir)
        # uncomment for debug
        # svc.cfg.report()

    # --------------------
    ## terminate.
    # reports overall rc
    #
    # @param verbose  flag for log verbosity
    # @return None
    def _term(self, verbose=True):
        # hack; some functions are not setting overallrc correctly
        if svc.gbl.overallrc == 0 and svc.gbl.rc != 0:
            svc.gbl.overallrc = svc.gbl.rc

        if verbose:
            svc.log.check(svc.gbl.overallrc == 0, f'{svc.gbl.tag}: overall rc={svc.gbl.overallrc}')
        sys.exit(svc.gbl.overallrc)



from dataclasses import dataclass


# --------------------
## holds singletons and global values
@dataclass
class svc:  # pylint: disable=invalid-name
    ## holds reference to Utils class
    utils = None

    ## holds reference to UtilsFs class; for file system functions
    utils_fs = None

    ## holds reference to UtilsPs class; for OS Process functions
    utils_ps = None

    ## holds reference to GenFiles class; for functions that generate specific files
    gen_files = None

    ## holds config variables
    cfg = None

    ## holds reference to logger
    log = None

    ## holds all global values
    @dataclass
    class Globals:
        ## the directory to hold output files
        outdir = 'out'
        ## holds logging tag
        tag = 'notset'
        ## holds the return code for the current operation
        rc = 0
        ## holds the overall return code for all the operations so far
        overallrc = 0
        ## indicates whether the logging should be verbose or not
        verbose = False

    ## holds reference to Globals
    gbl = Globals

    # --------------------
    ## abort the current script.
    # Note: do not use logging here since it may fail to write correctly
    #
    # @param msg  (optional) message to display
    # @return does not return
    @classmethod
    def abort(cls, msg='abort occurred, exiting'):
        import sys
        print('')
        print(f'ABRT {msg}')
        sys.stdout.flush()
        sys.stderr.flush()
        sys.exit(1)



import configparser
import os
from dataclasses import dataclass

from .os_specific import OsSpecific
from .svc import svc


# --------------------
## contains all configuration objects and values
@dataclass
class Cfg:
    ## the version string held in mod_dir_name/lib/version.json
    version = 'notset'
    ## the module name (with dashes)
    mod_name = 'notset'
    ## the local directory name for the module (with underscores)
    mod_dir_name = 'notset'
    ## flag indicating this is a python module (True) or an app (False)
    is_module = True
    ## the module/app type: public or private
    mod_type = 'notset'
    ## the tech for this project: python, cpp, arduino
    mod_tech = 'python'

    ## the license for the module
    license = 'notset'
    ## the license string for the classifier section
    classifier_license = 'notset'
    ## the url for the homepage link
    homepage_url = 'notset'
    ## the url for the download link
    download_url = 'notset'
    ## the author name
    author = 'notset'
    ## the contact email
    email = 'notset'
    ## the type of the description file used
    long_desc_type = 'notset'

    # ----
    ## attributes specific for do_publish script
    @dataclass
    class do_publish:  # pylint: disable=invalid-name
        ## list of projects to disallow
        disallow = []

    # ----
    ## attributes specific for do_lint script
    @dataclass
    class do_lint:  # pylint: disable=invalid-name
        ## if True, include tools directory in lint
        include_tools = False
        ## list of source directories to perform lint on
        src_dirs = []

    # ----
    ## attributes specific for do_doc script
    @dataclass
    class do_doc:  # pylint: disable=invalid-name
        ## the type: "Module" or "App"
        doxy_type = 'notset'
        ## the Doxyfile project description
        doxy_desc = 'notset'
        ## additional directories to exclude in the Doxyfile
        doxy_exclude = []
        ## additional directories to specifically include in the Doxyfile
        doxy_include = []

    # ----
    ## attributes specific for do_post_ver script
    @dataclass
    class do_post_ver:  # pylint: disable=invalid-name
        ## the destination directory for verification reports and files
        dst_dir = 'notset'
        ## directory to get verification html files
        src_dir = 'notset'
        ## web slug
        slug = 'notset'
        ## PyPi URL for python modules; use 'N/A' not a module
        pypi_url = 'notset'
        ## repo URL
        repo_url = 'notset'

    # ----
    ## attributes specific for do_ut script
    @dataclass
    class do_ut:  # pylint: disable=invalid-name
        ## set of directories for coverage
        cov_dirs = []

    # ----
    ## attributes specific for cpip
    @dataclass
    class cpip:  # pylint: disable=invalid-name
        ## the root directory of the cpip database
        root_dir = 'notset'
        ## the default server root url for pulls
        server_root = 'https://arrizza.com/web-cpip'
        ## the packages to use
        packages = []

    # == private from here

    ##  holds the root dir; usually '.'
    _root_dir = None
    ## flag indicates if the cfg values have been initialized
    _initialized = False

    # --------------------
    ## loads cfg values from xplat.cfg file
    #
    # @param root_dir    root directory to find .cfg file
    # @param init_once   (optional) if True, initialized only once
    # @return None
    @classmethod
    def load(cls, root_dir, init_once=True):
        if init_once and cls._initialized:
            return
        OsSpecific.init()
        cls._load_cfg_file(root_dir, 'xplat.cfg', init_once)

    # --------------------
    ## load the xplat configuration file
    #
    # @param root_dir   location of the cfg file
    # @param fname      name of the cfg file
    # @param init_once  initialize only once
    @classmethod
    def _load_cfg_file(cls, root_dir, fname, init_once):
        if init_once and cls._initialized:
            return

        cls._root_dir = root_dir
        path = os.path.join(root_dir, fname)
        config = configparser.ConfigParser()
        config.read(path)

        cls._load_xplat(config)
        cls._load_do_publish(config)
        cls._load_do_lint(config)
        cls._load_do_doc(config)
        cls._load_do_post_ver(config)
        cls._load_do_ut(config)
        cls._load_cpip(config)

        ## see class doc
        cls._initialized = True

    # --------------------
    ## loads values from xplat section.
    # Note: should not be any OS specific content in xplat section
    #
    # @param config  reference to ConfigParser object
    # @return None
    @classmethod
    def _load_xplat(cls, config):
        section = 'xplat'
        # uncomment to debug
        # print(f'==== section {section}')
        values = dict(config.items(section))

        for opt, orig_val in values.items():
            val = cls._parse_str(opt, orig_val)
            setattr(cls, opt, val)

            if opt == 'mod_name':
                if '_' in cls.mod_name:
                    svc.abort(f'mod_name should not have underscores: "{val}"')
                else:
                    setattr(cls, 'mod_dir_name', cls.mod_name.replace('-', '_'))

            elif opt == 'version':
                ## see class doc
                cls.long_version = cls.version.replace('.', '_')

    # --------------------
    ## loads values from do_publish section
    #
    # @param config  reference to ConfigParser object
    # @return None
    @classmethod
    def _load_do_publish(cls, config):
        cls._load_section(
            config,
            ## see class doc above
            cls.do_publish,
            'do_publish',
            []  # no lists options used
        )

    # --------------------
    ## loads values from do_lint section
    #
    # @param config  reference to ConfigParser object
    # @return None
    @classmethod
    def _load_do_lint(cls, config):
        cls._load_section(
            config,
            ## see class doc above
            cls.do_lint,
            'do_lint',
            ['src_dirs']
        )

    # --------------------
    ## loads values from do_doc section
    #
    # @param config  reference to ConfigParser object
    # @return None
    @classmethod
    def _load_do_doc(cls, config):
        cls._load_section(
            config,
            ## see class doc above
            cls.do_doc,
            'do_doc',
            ['doxy_exclude', 'doxy_include']
        )

        # add project type to Project_name in Doxyfile
        if cls.is_module:
            setattr(cls.do_doc, 'doxy_type', 'Module')
        else:
            setattr(cls.do_doc, 'doxy_type', 'App')

    # --------------------
    ## loads values from do_post_ver section
    #
    # @param config  reference to ConfigParser object
    # @return None
    @classmethod
    def _load_do_post_ver(cls, config):
        cls._load_section(
            config,
            ## see class doc above
            cls.do_post_ver,
            'do_post_ver',
            []  # no list opttions
        )

    # --------------------
    ## loads values from do_ut section
    #
    # @param config  reference to ConfigParser object
    # @return None
    @classmethod
    def _load_do_ut(cls, config):
        cls._load_section(
            config,
            ## see class doc above
            cls.do_ut,
            'do_ut',
            ['cov_dirs']
        )

    # --------------------
    ## loads values from cpip section
    #
    # @param config  reference to ConfigParser object
    # @return None
    @classmethod
    def _load_cpip(cls, config):
        cls._load_section(
            config,
            ## see class doc above
            cls.cpip,
            'cpip',
            ['packages']
        )

    # --------------------
    ## load a section from the xplat.cfg file
    #
    # @param config        the data from the xplat.cfg
    # @param section_ref   the data from the class cfg section above cls.xx
    # @param section_name  the name used in the xplat.cfg file
    # @param list_opts     the option names that should be parsed as lists
    # @return None
    @classmethod
    def _load_section(cls, config, section_ref, section_name, list_opts):
        valid_opts, values = cls._get_info_for(config, section_ref, section_name)
        for opt in valid_opts:
            value = cls._get_value_for(opt, values)
            if value is None:
                continue

            if opt in list_opts:
                val = cls._parse_list(opt, value)
            else:
                val = cls._parse_str(opt, value)
            setattr(section_ref, opt, val)

        # double check about extra values in the cfg.section
        for opt in values.keys():
            if opt not in valid_opts:
                opt2 = opt.split('.')
                # uncomment to debug
                # svc.log.dbg(f'opt2:{opt2} opt:{opt} {OsSpecific.os_name}')
                if opt2[1] not in OsSpecific.os_valid:
                    svc.log.warn(f'section {section_name} option has unrecognized OS: {opt}')
                if opt2[0] not in valid_opts:
                    svc.log.warn(f'section {section_name} has an invalid option: {opt}')

    # --------------------
    ## get the valid options and values for the given section in xplat.cfg
    #
    # @param config        the data from the xplat.cfg
    # @param section_ref   the data from the class cfg section above cls.xx
    # @param section_name  the name used in the xplat.cfg file
    # @return None
    @classmethod
    def _get_info_for(cls, config, section_ref, section_name):
        # the valid options for this section
        valid_opts = []
        for key in section_ref.__dict__.keys():
            if key.startswith('__'):
                continue
            valid_opts.append(key)

        # the values in the .cfg file for this section
        values = dict(config.items(section_name))
        # uncomment to debug
        # svc.log.dbg(f'@@@ {section_name}:')
        # svc.log.dbg(f'  - valid : {valid_opts}')
        # svc.log.dbg(f'  - values: {values}')
        return valid_opts, values

    # --------------------
    ## get the value for the given option from the values dict
    #
    # @param opt       the option to get
    # @param values    the list of values from the xplat.cfg file
    # @return the value is found, otherwise None
    @classmethod
    def _get_value_for(cls, opt, values):
        os_opt = f'{opt}.{OsSpecific.os_name}'
        if os_opt in values:
            # value for this OS
            value = values[os_opt]
            # uncomment to debug
            # svc.log.dbg(f'  - found : {os_opt}={value} (OS override)')
        elif opt in values:
            value = values[opt]
            # uncomment to debug
            # svc.log.dbg(f'  - found : {opt}={value}')
        else:
            value = None
        return value

    # --------------------
    ## Debug: report all values found in the cfg file
    #
    # @return None
    @classmethod
    def report(cls):
        # uncomment to debug
        svc.log.start('Cfg:')
        cls._report('cls', cls)
        cls._report('do_publish', cls.do_publish)
        cls._report('do_lint', cls.do_lint)
        cls._report('do_doc', cls.do_doc)
        cls._report('do_post_ver', cls.do_post_ver)
        cls._report('do_ut', cls.do_ut)
        cls._report('cpip', cls.cpip)

    # --------------------
    ## Debug: report all values found in a section
    #
    # @param sectname   section name to report
    # @param subclass   the location (a dataclass) of the attributes to report
    # @return None
    @classmethod
    def _report(cls, sectname, subclass):
        for opt, val in vars(subclass).items():
            if opt.startswith('__'):
                continue
            if callable(getattr(subclass, opt)):
                continue

            optname = f'{sectname}.{opt}'
            if opt == 'long_desc':
                optval = val[0]
            else:
                optval = val

            # sfx = ''
            # if val == 'notset':
            #     sfx = '; WARN value is not set'
            msg = f'{optname: <25}: [{type(optval).__name__: <4}]; {optval}'
            if val == 'notset':
                svc.log.warn(f'{msg}; value is not set')
            else:
                svc.log.line(msg)

    # --------------------
    ## substitue common values if found in a cfg string
    # e.g.: {mod_name}, {mod_dir_name}
    #
    # @param opt  the optional being used
    # @param val  the value to convert
    # @return None
    @classmethod
    def _sub_common(cls, opt, val):
        common_tags = {
            ## see doc above
            '{mod_name}': cls.mod_name,
            ## see doc above
            '{mod_dir_name}': cls.mod_dir_name,
            ## see doc above
            '{mod_type}': cls.mod_type,
        }

        for tag_name, tag_val in common_tags.items():
            if val.find(tag_name) != -1:
                val = val.replace(tag_name, tag_val)

        # report an invalid tag_name
        if '{' in val or '}' in val:
            svc.log.warn(f'{opt}: check for invalid tag_name: \'{val}\'')

        return val

    # --------------------
    ## parse the given string value and substitute any common values
    #
    # @param opt   the cfg name
    # @param val   the cfg value (a string)
    # @return the updated value
    @classmethod
    def _parse_str(cls, opt, val):  # pylint: disable=unused-argument
        val = val.strip("'")
        val = cls._sub_common(opt, val)

        # handle bool
        if '"' in val:
            svc.log.warn(f'{opt}: double quotes are not parsed out: \'{val}\'')
        elif val.lower() == 'true':
            val = True
        elif val.lower() == 'false':
            val = False

        # note: integers and floats are returned as strings

        # uncomment to debug
        # print(f'{opt: <25}: {val}')
        return val

    # --------------------
    ## parse the given list value and substitute any common values in each list item
    #
    # @param opt   the cfg name
    # @param val   the cfg value (a list)
    # @return the updated value
    @classmethod
    def _parse_list(cls, opt, val):
        val_list = []
        if '\n' in val:
            subvals = val.split('\n')
            for subval in subvals:
                if '[' in subval or ']' in subval:
                    svc.log.warn(f'{opt}: square brackets are not parsed out: \'{subval}\'')

                if subval == '':  # if empty don't add to list
                    svc.log.warn(f'{opt}: empty value in: \'{val}\'')
                else:
                    subval = cls._parse_str(opt, subval)
                    val_list.append(subval)
        elif ',' in val:
            subvals = val.split(',')
            for subval in subvals:
                subval = subval.strip(' ')
                if '[' in subval or ']' in subval:
                    svc.log.warn(f'{opt}: square brackets are not parsed out: \'{subval}\'')

                if subval == '':  # if empty don't add to list
                    svc.log.warn(f'{opt}: empty value in: \'{val}\'')
                else:
                    subval = cls._parse_str(opt, subval)
                    val_list.append(subval)
        elif val != '':  # if empty don't add to list
            val = cls._parse_str(opt, val)
            val_list.append(val)
        return val_list



import os
import re
import shutil

from .svc import svc


# --------------------
## perform the do_doc operation
class DoDoc:
    # --------------------
    ## do_doc mainline.
    #
    # @param make_pdf  gen the pdf (True) or not (False)
    # @return None
    def run(self, make_pdf=''):
        make_pdf = bool(make_pdf in ['', 'True'])

        svc.log.highlight(f'{svc.gbl.tag}: starting make_pdf={make_pdf}...')

        # ensure outdir is created
        svc.utils_fs.create_outdir()
        # ensure out/doc is created
        svc.utils_fs.clean_out_dir('doc')

        self._gen_doxyfile()
        self._gen_html()
        self._sort_warnings()
        if make_pdf:
            self._gen_pdf()
        else:
            svc.log.line(f'{svc.gbl.tag}: PDF not generated')

    # --------------------
    ## generate the Doxyfile cfg from tools/doxyfile_template
    #
    # @return None
    def _gen_doxyfile(self):
        svc.utils_fs.safe_delete_file('Doxyfile')

        # note: no subdir since doxygen will recurse
        src_dir = svc.utils.get_src_dir(svc.gbl.tag, '')

        # generate the Doxyfile file from doxyfile_template
        include_path = self._get_doxyfile_master()
        src = self._get_doxyfile_template()
        dst = os.path.join(svc.utils_fs.root_dir, 'Doxyfile')
        with open(dst, 'w', encoding='utf-8', newline='\n') as fp_dst:
            with open(src, 'r', encoding='utf-8') as fp_src:
                for line in fp_src:
                    while True:
                        m = re.search(r'\$\$([^$]+)\$\$', line)
                        if not m:
                            # uncomment to debug
                            # log.dbg(f'line: {line.strip()}')
                            if line.find('EXCLUDE ') != -1:
                                line = svc.utils_fs.slash_convert(line)
                            elif line.find('INPUT ') != -1:
                                line = svc.utils_fs.slash_convert(line)
                            elif line.find('@INCLUDE_PATH ') != -1:
                                line = svc.utils_fs.slash_convert(line)
                            elif line.find('OUTPUT_DIRECTORY ') != -1:
                                line = svc.utils_fs.slash_convert(line)
                            elif line.find('WARN_LOGFILE ') != -1:
                                line = svc.utils_fs.slash_convert(line)
                            fp_dst.write(line)
                            break

                        line = self._update_line(line, m.group(1), src_dir, include_path)

                for dirname in svc.cfg.do_doc.doxy_exclude:
                    path = svc.utils_fs.slash_convert(dirname)
                    line = f'EXCLUDE += {path}\n'
                    fp_dst.write(line)

                for dirname in svc.cfg.do_doc.doxy_include:
                    path = svc.utils_fs.slash_convert(dirname)
                    line = f'INPUT += {path}\n'
                    fp_dst.write(line)

        svc.gbl.overallrc += svc.gbl.rc

    # --------------------
    ## check if the user is using a local copy of doxyfile_master.
    #
    # @param line          the line to update
    # @param vrbl          the variable found in the original line
    # @param src_dir       the location of source files
    # @param include_path  the location of the doxyfile_master file
    # @return the updated line
    def _update_line(self, line, vrbl, src_dir, include_path):
        if vrbl == 'root_dir':
            val = str(svc.utils_fs.root_dir)
        elif vrbl == 'outdir':
            val = str(svc.gbl.outdir)
        elif vrbl == 'src_dir':
            val = src_dir
        elif vrbl in ['doxy_type', 'doxy_desc', 'doxy_exclude', 'doxy_include']:
            val = getattr(svc.cfg.do_doc, vrbl)
        elif vrbl == 'include_path':
            val = include_path
        else:
            val = getattr(svc.cfg, vrbl)

        # uncomment to debug
        # log.dbg(f'grp:{m.group(1)} line:{line.strip()}')
        line = line.replace(f'$${vrbl}$$', val)
        return line

    # --------------------
    ## check if the user is using a local copy of doxyfile_master.
    # otherwise default to the one in xplat_utils.
    #
    # @return the location of the doxyfile_master file
    def _get_doxyfile_master(self):
        path = os.path.join(svc.utils_fs.root_dir, 'tools', 'doxyfile_master')
        include_path = 'tools'
        if not os.path.isfile(path):
            include_path = 'tools/xplat_utils'
        return include_path

    # --------------------
    ## check if the user is using a local copy of doxyfile_template.
    # otherwise default to the one in xplat_utils.
    #
    # @return the location of the doxyfile_template file
    def _get_doxyfile_template(self):
        src = os.path.join(svc.utils_fs.root_dir, 'tools', 'doxyfile_template')
        if not os.path.isfile(src):
            src = os.path.join(svc.utils_fs.root_dir, 'tools', 'xplat_utils', 'doxyfile_template')
        return src

    # --------------------
    ## generate the doxygen HTML and latex content
    #
    # @return None
    def _gen_html(self):
        svc.log.highlight(f'{svc.gbl.tag}: running doxygen')
        svc.utils_ps.run_process('doxygen', use_raw_log=False)
        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: doxygen rc={svc.gbl.rc}')

        svc.gbl.overallrc += svc.gbl.rc

    # --------------------
    ## generate the doxygen PDF file from latex content
    #
    # @return None
    def _gen_pdf(self):
        svc.log.highlight(f'{svc.gbl.tag}: generate PDF')
        path = os.path.join(svc.gbl.outdir, 'doc', 'latex')
        log_file = os.path.join(svc.gbl.outdir, 'doxygen_latex.txt')
        svc.utils_ps.run_process('make', use_raw_log=False, working_dir=path, log_file=log_file)
        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: generate PDF rc={svc.gbl.rc}')
        if svc.gbl.rc != 0:
            svc.log.err(f'check the contents of {log_file}')
            svc.gbl.overallrc += svc.gbl.rc
            return

        pdf_path = os.path.join(svc.gbl.outdir, 'doc', 'latex', 'refman.pdf')
        pdf_path = pdf_path.replace('\\', '/')
        if os.path.isfile(pdf_path):
            svc.log.ok(f'{svc.gbl.tag}: see full output in: {log_file}')
        else:
            svc.log.err(f'{svc.gbl.tag}: PDF does not exist: {pdf_path}')
            svc.log.err(f'{svc.gbl.tag}: see full output in: {log_file}')

        dst = os.path.join(svc.gbl.outdir, f'{svc.cfg.mod_name}.pdf')
        shutil.copyfile(pdf_path, dst)

        svc.gbl.overallrc += svc.gbl.rc

    # --------------------
    ## sort the doxygen warnings in order of path to the file,
    # and the line number within in the file
    #
    # @return None
    def _sort_warnings(self):
        path = os.path.join(svc.gbl.outdir, 'doxygen.txt')
        warnings = 0
        lines = []
        with open(path, 'r', encoding='UTF-8') as fp:
            for line in fp:
                srch = f'{os.getcwd()}/(.*\n)'
                srch = srch.replace('\\', '/')
                m = re.search(srch, line)
                if m:
                    lines.append(m.group(1))
                else:
                    lines.append(line)

        lines = sorted(lines, key=self._get_warning_fields)

        with open(path, 'w', encoding='UTF-8', newline='\n') as fp:
            fp.writelines(lines)

        for line in lines:
            if line != '':
                warnings += 1
                svc.log.raw(line.strip())
        if warnings != 0:
            svc.gbl.rc += warnings

        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: found {warnings} warnings')
        svc.gbl.overallrc += svc.gbl.rc

    # --------------------
    ## get fields from a doxygen warning line.
    # used to sort them
    #
    # @param line  the current line
    # @return tuple with file path and line number within the file
    def _get_warning_fields(self, line):
        m = re.search(r'([^:]+):(\d+):', line)
        if m:
            return m.group(1), int(m.group(2))

        return line, ''



import os
import typing

from .svc import svc


# ---------------------
## common functions;
class UtilsCommon:
    # ---------------------
    ## get the top directory of src files.
    # aborts if the directory does not exist
    #
    # @param tag     for logging purposes
    # @param subdir  optional subdirectory
    # @return the relative directory path to the source
    @staticmethod
    def get_src_dir(tag, subdir) -> typing.Optional[str]:
        path = ''
        if svc.cfg.mod_tech in ['python']:
            if svc.cfg.is_module:
                mod_dir = os.path.join(svc.utils_fs.root_dir, svc.cfg.mod_dir_name)
                if subdir:
                    path = os.path.join(str(mod_dir), 'lib', subdir)
                else:
                    path = os.path.join(str(mod_dir), 'lib')
            else:
                if subdir:
                    path = os.path.join(svc.utils_fs.root_dir, 'lib', subdir)
                else:
                    path = os.path.join(svc.utils_fs.root_dir, 'lib')
        elif svc.cfg.mod_tech in ['cpp', 'arduino']:
            if svc.cfg.is_module:
                if subdir:
                    path = os.path.join(svc.utils_fs.root_dir, 'lib', subdir)
                else:
                    path = os.path.join(svc.utils_fs.root_dir, 'lib')
            else:
                if subdir:
                    path = os.path.join(svc.utils_fs.root_dir, 'src', subdir)
                else:
                    path = os.path.join(svc.utils_fs.root_dir, 'src')
        else:
            svc.gbl.rc = 1
            svc.abort(f'{tag}: unknown tech: {svc.cfg.mod_tech}')  # okay to abort; tech should always be defined

        if path and not os.path.isdir(path):
            svc.gbl.rc = 1
            svc.abort(f'{tag}: dir does not exist: {path}')  # okay to abort; directory should always exist

        # uncomment to debug
        # print(f'DBG @@@ {tag} path:{path}')

        return str(path)

    # --------------------
    ## return the module name for this given arg
    #  if var is invalid, AttributeError is thrown
    #
    # @param var   the variable to retrieve
    # @return the module name or an error message
    @staticmethod
    def get_cfg(var):
        return getattr(svc.cfg, var)

    # --------------------
    ## return the Coverage options needed for the module/app.
    # note: these are for python only; see do_coverage for cpp and other techs
    #
    # @return coverage options needed for pytest (python)
    @staticmethod
    def get_cov_opts():
        opts = ''
        # the directory to cover
        if svc.cfg.is_module:
            opts += f'--cov={svc.cfg.mod_dir_name}/lib '
        else:  # an app
            opts += '--cov=lib '

        opts += '--cov-report= '  # the type of report; default is HTML
        opts += '--cov-branch '  # branch coverage
        opts += '--cov-config=setup.cfg '  # other cfg is in setup.cfg
        opts += '--cov-append'  # append to coverage content; up to caller to clear it
        return opts



import glob
import os

from .svc import svc


# --------------------
## perform the do_lint operation
class DoLint:
    # --------------------
    ## do_lint mainline.
    #
    # @param tool   for python: pylint (default) or ruff; for cpp: cppcheck or clang-tidy(default)
    # @return None
    def run(self, tool):
        if svc.cfg.mod_tech == 'python' and not tool:
            tool = 'pylint'
        elif svc.cfg.mod_tech in ['cpp', 'arduino'] and not tool:
            tool = 'cppcheck'

        svc.log.highlight(f'{svc.gbl.tag}: starting tech:{svc.cfg.mod_tech} tool:{tool}...')
        if svc.cfg.mod_tech == 'python':
            self._run_python(tool)
        elif svc.cfg.mod_tech in ['cpp', 'arduino']:
            # arduino defaults to same as C++
            self._run_cpp(tool)
        else:
            svc.gbl.rc += 1
            svc.log.err(f'unknown tech:{svc.cfg.mod_tech}, use one of: python, cpp, arduino')

        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: pylint rc={svc.gbl.rc}')

    # --------------------
    ## run pylint
    #
    # @param tool  the tool to use; valid values depend on current tech
    # @return None
    def _run_python(self, tool):
        # always run pycodestyle
        svc.log.highlight(f'{svc.gbl.tag}: running pycodestyle')
        svc.utils_ps.run_process('pycodestyle', use_raw_log=True)
        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: pycodestyle rc={svc.gbl.rc}')

        if tool == 'pylint':
            svc.log.highlight(f'{svc.gbl.tag}: running pylint')
            dirs = self._get_dirs()
            rc_path = self._find_pylintrc()
            cmd = f'pylint --rcfile={rc_path} {dirs}'
            svc.utils_ps.run_process(cmd, use_raw_log=True)
        elif tool == 'ruff':
            svc.log.highlight(f'{svc.gbl.tag}: running ruff')
            toml_path = self._find_ruff_toml()
            cmd = f'ruff check --config {toml_path}'
            svc.utils_ps.run_process(cmd, use_raw_log=True)
        else:
            svc.gbl.rc += 1
            svc.log.err(f'unknown tool:"{tool}", use one of: pylint, ruff')

    # --------------------
    ## get location of pylint.rc file.
    # Note: for pylint (python) only
    #
    # @return path to pylint.rc
    def _find_pylintrc(self):
        rc_path = os.path.join(svc.utils_fs.root_dir, 'tools', 'pylint.rc')
        if not os.path.isfile(rc_path):
            rc_path = os.path.join(svc.utils_fs.root_dir, 'tools', 'xplat_utils', 'pylint.rc')
        return rc_path

    # --------------------
    ## get location of ruff.toml file.
    # Note: for ruff (python) only
    #
    # @return path to ruff.toml
    def _find_ruff_toml(self):
        toml_path = os.path.join(svc.utils_fs.root_dir, 'tools', 'ruff.toml')
        if not os.path.isfile(toml_path):
            toml_path = os.path.join(svc.utils_fs.root_dir, 'tools', 'xplat_utils', 'ruff.toml')
        return toml_path

    # --------------------
    ## get list of directories to pylint.
    # Note: for pylint (python) only
    #
    # @return list of directories
    def _get_dirs(self):
        modules = ''
        if svc.cfg.is_module:
            modules += f'{svc.cfg.mod_dir_name} '
        else:
            modules += 'lib '

        files = glob.glob('*.py')
        if files:
            modules += '*.py '

        if svc.cfg.do_lint.include_tools:
            svc.log.line(f'{svc.gbl.tag}: including "tools" directory')
            modules += 'tools '
            path = os.path.join('tools', 'xplat_utils')
            modules += f'{path} '

        # include ver and ut only if the directories exist
        if os.path.isdir('ver'):
            modules += 'ver '
        if os.path.isdir('ut'):
            modules += 'ut '

        return modules

    # --------------------
    ## run cppcheck or clang-tidy
    #
    # @param tool  which lint tool to use: clang-tidy(default), cppcheck
    # @return None
    def _run_cpp(self, tool):
        if tool == 'clang-tidy':
            self._run_cpp_clang_tidy()
        elif tool == 'cppcheck':
            self._run_cpp_check()
        else:
            svc.gbl.rc += 1
            svc.log.err(f'unknown tool:"{tool}", use one of: clang-tidy, cppcheck')

    # --------------------
    ## run cppcheck
    #
    # @return None
    def _run_cpp_check(self):
        modules = ''
        # location of source files being covered
        for src_dir in svc.cfg.do_lint.src_dirs:
            modules += f'{src_dir} '  # pylint: disable=consider-using-join

        svc.log.highlight(f'{svc.gbl.tag}: run cppcheck on: {modules}')
        cmd = 'cppcheck '
        cmd += '--enable=all '
        cmd += '--inconclusive  '  # _         report errors even if inconclusive
        cmd += '--max-ctu-depth=5 '  # _       minimize max-cross-translation unit checks
        # cmd += '--language=c++ ' # _         not needed
        cmd += '--inline-suppr '  # _          allow inline suppression e.g "// cppcheck-suppress unusedFunction"
        cmd += '--suppress=checkersReport '  # suppress comment to gen checkers report
        if svc.cfg.mod_tech == 'arduino':
            cmd += '--platform=avr8 '  # _     arduino uses avr8 compiler
            cmd += '--suppress=missingIncludeSystem '  # ignore these for arduino
        else:
            cmd += '--platform=native '  # _   use native compiler spec
        cmd += modules
        svc.utils_ps.run_process(cmd, use_raw_log=False)
        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: cppcheck rc={svc.gbl.rc}')

    # --------------------
    ## run clang-tidy
    #
    # @return None
    def _run_cpp_clang_tidy(self):
        # skip regen for Makefile projects
        # TODO delete when cmake is deprecated
        if os.path.isfile('CMakeLists.txt'):
            svc.log.highlight(f'{svc.gbl.tag}: cmake refresh compile_commands')
            cmd = 'cmake -DCMAKE_EXPORT_COMPILE_COMMANDS=ON cmake-build-debug/'
            svc.utils_ps.run_process(cmd, use_raw_log=False)
            svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: compile_commands rc={svc.gbl.rc}')

        modules = ''
        # location of source files being covered
        for src_dir in svc.cfg.do_lint.src_dirs:
            modules += f'{src_dir}/* '  # pylint: disable=consider-using-join

        svc.log.highlight(f'{svc.gbl.tag}: run clang-tidy on: {modules}')
        cmd = 'clang-tidy '
        cmd += '-p=cmake-build-debug/compile_commands.json '  # TODO delete when cmake is deprecated
        cmd += '--quiet '
        cmd += '--header-filter=.* '
        cmd += modules
        svc.utils_ps.run_process(cmd, use_raw_log=False)
        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: clang-tidy rc={svc.gbl.rc}')



import os
import re
import shutil
from pathlib import Path

from .os_specific import OsSpecific
from .svc import svc


# --------------------
## holds all file system related utility functions
class UtilsFs:
    # --------------------
    ## constructor
    def __init__(self):
        OsSpecific.init()
        ## holds the root directory for all files and directories
        self._root_dir = None

    # --------------------
    ## convert path to appropriate slash for the OS
    #
    # @param path  the path to convert
    # @return the convert path
    def slash_convert(self, path):
        if OsSpecific.os_name == 'win':
            return path.replace('/', '\\')
        return path.replace('\\', '/')

    # --------------------
    ## convert path with leading C:\ to msys2 compatible
    #
    # @param path  the path to convert
    # @return the convert path
    def full_convert(self, path):
        if OsSpecific.os_name == 'win':
            path = path.replace('\\', '/')
            path = path.replace('//', '/')
            m = re.search(r'^(.):/(.*)', path)
            if m:
                path = f'/{m.group(1)}/{m.group(2)}'

        return path

    # --------------------
    ## returns the root directory whether in a module in venv or locally
    #
    # @return the root dir
    @property
    def root_dir(self):
        if self._root_dir is None:
            self._root_dir = str(Path('..').parent)

        return self._root_dir

    # --------------------
    ## create the out directory.
    # return the outdir as an absolute path
    #
    # @return the module name or an error message
    def create_outdir(self):
        path = self.convert_relative_to_abs(svc.gbl.outdir)
        self.safe_create_dir(path)
        return str(path)

    # --------------------
    ## convert a relative path to an absolute path, e.g.
    # * ~/out    => /home/xx/out
    # * ./out    => /home/xx/projects/proj1/out
    # * out      => /home/xx/projects/proj1/out
    #
    # @param path  the path to convert
    # @return the converted path
    def convert_relative_to_abs(self, path):
        path1 = os.path.expanduser(path)
        path2 = os.path.abspath(path1)
        return path2

    # --------------------
    ## delete and then recreate the given subdirectory in "${OUT_DIR}"
    #
    # @param dst      the subdirectory to delete and recreate
    # @param verbose  logging verbosity
    # @return None
    def clean_out_dir(self, dst, verbose='quiet'):
        path1 = self.convert_relative_to_abs(svc.gbl.outdir)
        path = os.path.join(path1, dst)
        self.safe_delete_tree(path, verbose)
        self.safe_create_dir(path)

    # --------------------
    ## deletes a file in the given path in the root directory
    #
    # @param dst_path  the file to delete
    # @param verbose  logging verbosity
    # @return None
    def safe_delete_file(self, dst_path, verbose='quiet'):
        path = os.path.join(self.root_dir, dst_path)

        if not os.path.isfile(path):
            if verbose in ['full']:
                svc.log.ok(f'{"rm: file does not exist": <25}: {path}')
            return

        try:
            # uncomment to debug
            # svc.log.dbg(f'rm file: {path}')
            os.remove(path)
            if verbose in ['full', 'min']:
                svc.log.ok(f'{"rm: file ": <25}: {path}')
        except OSError as excp:  # pragma: no cover
            # coverage: needs sudo access to cause a delete failure
            svc.log.err(f'{"failed to delete file": <25}: {excp.filename}, {excp.strerror}')
            svc.gbl.rc += 1

    # --------------------
    ## deletes a sub-directory in the root dir
    #
    # @param dst_path  the subdirectory to delete
    # @param verbose  logging verbosity
    # @return None
    def safe_delete_tree(self, dst_path, verbose='quiet'):
        path = os.path.join(self.root_dir, dst_path)

        if not os.path.isdir(path):
            if verbose in ['full']:
                svc.log.ok(f'{"rm: dir  does not exist": <25}: {path}')
            return

        try:
            shutil.rmtree(path)
            if verbose in ['full', 'min']:
                svc.log.ok(f'{"rm: dir": <25}: {path}')
        except OSError as excp:  # pragma: no cover
            # coverage: needs sudo access to cause a delete failure
            svc.log.err(f'{"failed to delete dir": <25}: {excp.filename}, {excp.strerror}')
            svc.gbl.rc += 1

    # --------------------
    ## create the given subdirectory if not already existing
    #
    # @param dst   the subdirectory to create
    # @return None
    def safe_create_dir(self, dst):
        path = os.path.join(self.root_dir, dst)
        os.makedirs(path, exist_ok=True)

    # --------------------
    ## create the given global directory if not already existing
    #
    # @param path   the directory to create
    # @return None
    def safe_global_dir(self, path):
        path = os.path.expanduser(path)
        if os.path.isdir(path):
            svc.log.ok(f'already exists: {path}')
            return

        Path(path).mkdir(parents=True, exist_ok=True)
        if os.path.isdir(path):
            svc.log.ok(f'created: {path}')
        else:  # pragma: no cover
            # coverage: needs sudo access to cause a delete failure
            svc.log.err(f'failed to mkdir: {path}')

    # --------------------
    ## create the file from src to the dst_dir with the new file name
    #
    # @param src        the full path to the file to copy
    # @param dst_dir    the directory to copy to
    # @param dst_file   the new name of the file
    # @return None
    def safe_copy_file(self, src, dst_dir, dst_file):
        dst_dir = os.path.expanduser(dst_dir)
        if not os.path.isdir(dst_dir):
            svc.log.err(f'cannot find: {dst_dir}')
            svc.gbl.rc += 1
            return

        dst = os.path.join(dst_dir, dst_file)
        shutil.copyfile(src, dst)



import argparse
import os
import sys
import time

from falcon_logger.lib.logger import FalconLogger
from normal_logger import NormalLogger
from rsyslog_logger import RsyslogLogger
from stdout_logger import StdoutLogger


# -------------------
class App:
    # -------------------
    def __init__(self):
        self._system = None
        self._num_lines = 100
        self._log = None

    # -------------------
    def init(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('system')
        parser.add_argument('--numlines', default=100, type=int, required=False)
        args = parser.parse_args()
        # print(args)

        self._system = args.system
        self._num_lines = args.numlines

        self._log = None
        if self._system == 'stdout':
            self._log = StdoutLogger()
        elif self._system == 'falcon':
            self._log = FalconLogger(os.path.join('out', 'sample_falcon.log'))
        elif self._system == 'falcon2':
            self._log = FalconLogger(None)
        elif self._system == 'normal':
            self._log = NormalLogger()
        elif self._system == 'rsyslog':
            self._log = RsyslogLogger()
        else:
            print(f'unknown system: {self._system}, choose stdout, falcon, normal, rsyslog')
            sys.exit(1)

    # -------------------
    def term(self):
        pass

    # -------------------
    def run(self):
        start_time = time.time()
        for i in range(self._num_lines):
            self._log.debug(f'{i}: test')

        if self._log is not None:
            self._log.term()

        end_time = time.time()
        elapsed = round((end_time - start_time) * 1_000, 1)
        print(f'{self._system}: total time: {elapsed} ms')


# -------------------
def main():
    app = App()
    app.init()
    app.run()
    app.term()


main()



import json
import os

from .gen_build_info import GenBuildInfo
from .gen_constants import GenConstants
from .svc import svc


# --------------------
## perform gen files operations
class GenFiles:
    # --------------------
    ## generate all files
    #
    # @param dogen    ''      => no files
    #                 'dogen' => python : gen version.json and build_info.py
    #                 'dogen' => cpp,ard: gen version.h and build_info.txt
    #                 'dogen1'=> python : gen version.py
    #                 'dogen1'=> cpp,ard: gen version.h
    #                 'dogen2'=> python: gen version.py and build_info.py
    #                 'dogen2'=> cpp   : gen version.h and build_info.txt
    # @param subdir  (optional) subdirectory to place the files
    # @return None
    def all(self, dogen='', subdir=''):
        if svc.cfg.mod_tech not in ['python', 'cpp', 'arduino']:
            svc.gbl.rc += 1
            svc.log.err(f'unknown tech:{svc.cfg.mod_tech}, use one of: python, cpp or arduino')
        elif dogen not in ['', 'dogen', 'gen_both', 'gen_constants', 'gen_build_info']:
            svc.gbl.rc += 1
            svc.log.err(f'unknown dogen:{dogen}, use one of: "" or gen_both, gen_constants, gen_build_info')
        elif dogen in ['', 'dogen']:
            self._do_gen(subdir)
        elif dogen == 'gen_both':
            self._do_gen_both(subdir)
        elif dogen == 'gen_constants':
            self._do_gen_constants(subdir)
        else:  # dogen == 'gen_build_info':
            self._do_gen_build_info(subdir)

    # --------------------
    ## run do_gen option
    #
    # @param subdir the subdirectory in the source root directory
    # @return None
    def _do_gen(self, subdir):
        if svc.cfg.mod_tech == 'python':
            self._version_json(subdir)
        else:  # svc.cfg.mod_tech in ['cpp', 'arduino']
            self._version_h(subdir)
        self._do_gen_build_info(subdir)

    # --------------------
    ## run gen_both option
    #
    # @param subdir the subdirectory in the source root directory
    # @return None
    def _do_gen_both(self, subdir):
        self._do_gen_constants(subdir)
        self._do_gen_build_info(subdir)

    # --------------------
    ## run gen_constants option
    #
    # @param subdir the subdirectory in the source root directory
    # @return None
    def _do_gen_constants(self, subdir):
        if svc.cfg.mod_tech == 'python':
            impl = GenConstants()
            impl.run(subdir)
        else:  # svc.cfg.mod_tech in ['cpp', 'arduino']
            self._version_h(subdir)

    # --------------------
    ## run gen_build_info option
    #
    # @param subdir the subdirectory in the source root directory
    # @return None
    def _do_gen_build_info(self, subdir):
        if svc.cfg.mod_tech == 'python':
            self._build_info_file(subdir)
        else:  # svc.cfg.mod_tech in ['cpp', 'arduino']
            self._build_info_txt(subdir)

    # --------------------
    ## generate the version.json file in the module/app directory
    #
    # @param subdir the subdirectory in the source root directory
    # @return None
    def _version_json(self, subdir):
        fname = 'version.json'
        path = os.path.join(svc.utils.get_src_dir('version_json', subdir), fname)

        version = {'version': svc.cfg.version}
        with open(path, 'w', encoding='utf-8', newline='\n') as fp:
            json.dump(version, fp, indent=4)

        if os.path.isfile(path):
            svc.gbl.rc = 0
        else:
            svc.gbl.rc = 1

        self._report_rc(fname)

    # ---------------------
    ## generate version.h for cpp/arduino
    #
    # @param subdir the subdirectory in the source root directory
    # @return None
    def _version_h(self, subdir):  # pylint: disable=unused-argument
        fname = 'version.h'
        path = os.path.join(svc.utils.get_src_dir('version_h', subdir), fname)

        with open(path, 'w', encoding='utf-8', newline='\n') as fp:
            fp.write('#ifndef VERSION_H\n')
            fp.write('#define VERSION_H\n')
            fp.write(f'static const char version[] = "{svc.cfg.version}";\n')
            fp.write('#endif //VERSION_H\n')

        if os.path.isfile(path):
            svc.gbl.rc = 0
        else:
            svc.gbl.rc = 1

        self._report_rc(fname)

    # --------------------
    ## generate the build_info.py file in the module/app source directory
    #
    # @param subdir the subdirectory in the source root directory
    # @return None
    def _build_info_file(self, subdir):
        fname = 'build_info.py'
        path = os.path.join(svc.utils.get_src_dir('build_info_file', subdir), fname)
        self._gen_build_info(path)
        self._report_rc(fname)

    # ---------------------
    ## generate build_info.txt file
    #
    # @param subdir the subdirectory in the source root directory
    # @return None
    def _build_info_txt(self, subdir):
        fname = 'build_info.txt'
        path = os.path.join(svc.utils.get_src_dir('build_info_txt', subdir), fname)
        self._gen_build_info(path)
        self._report_rc(fname)

    # ---------------------
    ## generate build_info file at the given path
    #
    # @param path the location to gen the file into
    # @return None
    def _gen_build_info(self, path):
        binfo = GenBuildInfo()
        binfo.init(path)
        svc.gbl.rc = binfo.gen(svc.cfg.version)
        binfo.term()

    # ---------------------
    ## report rc to generate given filename
    #
    # @param fname  the file generated
    # @return None
    def _report_rc(self, fname):
        msg = f'gen: {fname} rc={svc.gbl.rc}'
        if not svc.gbl.verbose:
            svc.log.line(msg)
        elif svc.gbl.rc == 0:
            svc.log.ok(msg)
        else:
            svc.log.warn(msg)
        svc.gbl.overallrc += svc.gbl.rc



from pathlib import Path

from setuptools import find_packages
from setuptools import setup

print('     setup: version:  v$$version$$')
print('     setup: module :  $$mod_dir_name$$')

# @formatter:off
setup(
    description='module for faster python logging',
    # TODO update as needed
    keywords=['gui', 'test', 'verification'],
    install_requires=[
        'medver-pytest',
        'pytest',
        # TODO update as needed
        # 'socket-oneline',
    ],
    classifiers=[
        # TODO update as needed
        # Choose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'Programming Language :: Python :: 3.9',
        '$$classifier_license$$',
    ],

    # common attributes from here on
    name='$$mod_name$$',
    packages=find_packages(include='$$root_dir$$/$$mod_dir_name$$*', ),
    include_package_data=True,
    exclude_package_data={'$$root_dir$$/$$mod_dir_name$$/lib': ['.gitignore']},
    version='$$version$$',
    license='$$license$$',
    long_description=(Path(__file__).parent / 'README.md').read_text(),
    long_description_content_type='$$long_desc_type$$',
    author='$$author$$',
    author_email='$$email$$',
    url='$$homepage_url$$',
    download_url='$$download_url$$',
)
# @formatter:on

print('     setup: done')



from pathlib import Path

from setuptools import find_packages
from setuptools import setup

print('     setup: version:  v0.0.4')
print('     setup: module :  falcon_logger')

# @formatter:off
setup(
    description='module for faster python logging',
    # TODO update as needed
    keywords=['gui', 'test', 'verification'],
    install_requires=[
        'medver-pytest',
        'pytest',
        # TODO update as needed
        # 'socket-oneline',
    ],
    classifiers=[
        # TODO update as needed
        # Choose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
    ],

    # common attributes from here on
    name='falcon-logger',
    packages=find_packages(include='./falcon_logger*', ),
    include_package_data=True,
    exclude_package_data={'./falcon_logger/lib': ['.gitignore']},
    version='0.0.4',
    license='MIT',
    long_description=(Path(__file__).parent / 'README.md').read_text(),
    long_description_content_type='text/markdown',
    author='J. Arrizza',
    author_email='cppgent0@gmail.com',
    url='https://bitbucket.org/arrizza-public/falcon-logger/src/master',
    download_url='https://bitbucket.org/arrizza-public/falcon-logger/get/master.zip',
)
# @formatter:on

print('     setup: done')



import glob
import os
import re

from .svc import svc


# --------------------
## perform the do_coverage operation
class DoCoverage:
    # --------------------
    ## do_lint mainline.
    #
    # @param action  the action to take: reset, gen
    # @return None
    def run(self, action):
        svc.log.highlight(f'{svc.gbl.tag}: starting {action}...')

        tech = svc.cfg.mod_tech

        if action not in ['gen', 'reset']:
            svc.gbl.rc += 1
            svc.log.err(f'{svc.gbl.tag}: unknown action:{action}, use one of: reset, gen')
        elif tech not in ['python', 'cpp', 'arduino']:
            svc.gbl.rc += 1
            svc.log.err(f'{svc.gbl.tag}: unknown tech:{tech}, use one of: python, cpp, arduino')
        elif tech == 'python':
            if action == 'reset':
                svc.log.warn(f'{svc.gbl.tag}: {action}: not applicable to python tech, skipping')
            else:
                self._python_gen_coverage()
                self._python_report_summary()
        elif tech in ['cpp', 'arduino']:
            if action == 'reset':
                self._cpp_reset_coverage()
            else:  # gen
                self._cpp_gen_coverage()

        svc.log.check(svc.gbl.rc == 0, f'{svc.gbl.tag}: tech:{tech} action:{action} rc={svc.gbl.rc}')

    # --------------------
    ## generate coverage from pylint
    #
    # @return None
    def _python_gen_coverage(self):
        cmd = 'coverage html --rcfile setup.cfg'
        rc, out = svc.utils_ps.run_cmd(cmd)
        svc.gbl.rc += rc
        svc.log.line(f'{svc.gbl.tag}: {out}')

    # --------------------
    ## report on current coverage values
    # assumes: index.html has a known format
    #
    # @return None
    def _python_report_summary(self):
        path = os.path.join(svc.gbl.outdir, 'coverage', 'index.html')
        with open(path, 'r', encoding='utf-8') as fp:
            lines = fp.readlines()

            names = ['statements', 'missing', 'excluded', 'branches', 'partial', 'coverage']
            in_details = False
            detail_num = 0
            for line in lines:
                m = re.search(r'class="pc_cov">(.*)<', line)
                if m:
                    svc.log.line(f'{svc.gbl.tag}: Total: {m.group(1)}')
                    continue

                if not in_details:
                    m = re.search(r'<tfoot>', line)
                    if m:
                        in_details = True
                        detail_num = 0
                else:
                    m = re.search(r'<td>(\d+)</td>', line)
                    if m:
                        svc.log.line(f'{svc.gbl.tag}:    {names[detail_num]: <10}: {m.group(1): >5}')
                        detail_num += 1

    # --------------------
    ## reset gcov counters by deleting cmake-build-debug/*.gcda files
    # Note: if you delete .gcno => you have to recreate cmake-build-debug
    def _cpp_reset_coverage(self):
        path = 'cmake-build-debug'
        files = glob.glob(f'{path}/**/*.gcda', recursive=True)
        for file in files:
            svc.utils_fs.safe_delete_file(file, verbose=False)

    # --------------------
    ## generate coverage using gcovr.
    # see https://gcovr.com/en/4.2/guide.html
    # * Note: gcovr --gcov-executable gcov-8
    # * Note: gcovr --gcov-executable "llvm-cov gcov" # For CLANG
    #
    # @return None
    def _cpp_gen_coverage(self):
        covdir = f'{svc.gbl.outdir}/coverage'
        svc.utils_fs.clean_out_dir('coverage', verbose=False)

        report_page = f'{covdir}/ut.html'

        cmd = ('gcovr --html-details '  # show individual source files
               '-r cmake-build-debug '  # default to debug for unit test coverage
               '--sort=uncovered-percent '  # sort source files based on percentage uncovered lines
               '--print-summary '  # print summary to stdout
               f'-o {report_page} '  # location of report main page
               )

        # all directories named in xplat.do_ut
        for cov_dir in svc.cfg.do_ut.cov_dirs:
            cmd += f'--filter {cov_dir} '  # location of source files being covered

        svc.log.line(f'{svc.gbl.tag}: summary report:')
        svc.utils_ps.run_process(cmd, use_raw_log=False)
        svc.log.line(f'{svc.gbl.tag}: report is located in: {report_page}')



class BuildInfo:  # pylint: disable=too-few-public-methods
    python_version = '3.10.15'
    os_name = 'posix:ubuntu'
    version = '0.0.4'
    git_sha = 'da562b7c91033f5d29f4c3e04479eec01e215377'
    git_branch = 'master'
    git_uncommitted = [
    ]
    git_unpushed = [
    ]



import glob
import os

from .svc import svc


# --------------------
## perform the do_clean operation
class DoClean:
    # --------------------
    ## do_clean mainline.
    #
    # @param subdir   subdirectory if any
    # @param depth    how much to clean: lite, full;  '' defaults to lite
    # @return None
    def run(self, subdir='', depth=''):
        verbose = 'min'
        if depth == '':
            depth = 'lite'

        svc.log.highlight(f'{svc.gbl.tag}: starting tech:{svc.cfg.mod_tech} '
                          f'module:{svc.cfg.is_module} depth:{depth}...')

        svc.utils_fs.safe_delete_tree(svc.gbl.outdir, verbose=verbose)
        svc.utils_fs.safe_delete_file('Doxyfile', verbose=verbose)
        svc.utils_fs.safe_delete_file('.coverage', verbose=verbose)
        svc.utils_fs.safe_delete_tree('dist', verbose=verbose)
        svc.utils_fs.safe_delete_tree('.pytest_cache', verbose=verbose)

        if svc.cfg.mod_tech in ['python']:
            self._del_python(verbose, subdir)
        elif svc.cfg.mod_tech in ['cpp', 'arduino']:
            self._del_cpp(verbose, subdir)

        if depth == 'full':
            svc.utils_fs.safe_delete_tree('venv', verbose=verbose)

            # must be after venv
            self._delete_cache(verbose)

        # note: no return code

    # --------------------
    ## delete python specific directories and files
    #
    # @param verbose  the logging verbosity
    # @param subdir   the subdirectory in the source root directory
    # @return None
    def _del_python(self, verbose, subdir):
        if svc.cfg.is_module:
            if subdir:
                filedir = os.path.join(svc.cfg.mod_dir_name, 'lib', subdir)
            else:
                filedir = os.path.join(svc.cfg.mod_dir_name, 'lib')
        else:
            if subdir:
                filedir = os.path.join('lib', subdir)
            else:
                filedir = os.path.join('lib')

        if svc.cfg.is_module:
            svc.utils_fs.safe_delete_tree(os.path.join(f'{svc.cfg.mod_dir_name}.egg-info'), verbose=verbose)
            svc.utils_fs.safe_delete_file(os.path.join('setup.py'), verbose=verbose)
            svc.utils_fs.safe_delete_file(os.path.join('MANIFEST.in'), verbose=verbose)
            svc.utils_fs.safe_delete_file(os.path.join(filedir, 'build_info.py'), verbose=verbose)
            svc.utils_fs.safe_delete_file(os.path.join(filedir, 'version.json'), verbose=verbose)
            # don't delete constants_version.py
        else:  # an app
            svc.utils_fs.safe_delete_file(os.path.join(filedir, 'build_info.py'), verbose=verbose)
            svc.utils_fs.safe_delete_file(os.path.join(filedir, 'version.json'), verbose=verbose)
            # don't delete constants_version.py

    # --------------------
    ## delete cpp specific directories and files
    #
    # @param verbose  the logging verbosity
    # @param subdir   the subdirectory in the source root directory
    # @return None
    def _del_cpp(self, verbose, subdir):
        svc.utils_fs.safe_delete_tree('cmake-build-debug', verbose=verbose)
        svc.utils_fs.safe_delete_tree('cmake-build-release', verbose=verbose)
        svc.utils_fs.safe_delete_tree('debug', verbose=verbose)
        svc.utils_fs.safe_delete_tree('release', verbose=verbose)

        if svc.cfg.is_module:
            if subdir:
                filedir = os.path.join('lib', subdir)
            else:
                filedir = os.path.join('lib')
        else:
            if subdir:
                filedir = os.path.join('src', subdir)
            else:
                filedir = os.path.join('src')

        svc.utils_fs.safe_delete_file(os.path.join(filedir, 'build_info.txt'), verbose=verbose)
        # don't delete version.h

    # --------------------
    ## recursively deletes __pycache__ subdirectories in the root_dir.
    #
    # @param verbose  logging verbosity: 'full' or ''
    # @return None
    def _delete_cache(self, verbose):
        folders = glob.glob('**/__pycache__', recursive=True)
        # uncomment to debug
        # svc.log.dbg(f'{folders}')

        if not folders:
            return

        # at least one cache folder exists
        for folder in folders:
            # uncomment to debug
            # svc.log.dbg(f'rmtree {folder}')
            if folder.startswith('./venv'):
                continue
            svc.utils_fs.safe_delete_tree(folder, verbose=verbose)
